import numpy as np
import matplotlib.pyplot as plt
import time
from numba import jit
from scipy import optimize as op
import math

##########################################################
##########################################################
###                                                    ###
###   Just press play to create all plots in thesis!   ###
###                                                    ###
##########################################################
##########################################################

t0 = time.clock() #Timing the program

#################################
#################################
###                           ###
###   Defining functions      ###
###                           ###
#################################
#################################


#################################
#################################
###                           ###
###   Non-relativistic case   ###
###                           ###
#################################
#################################

#########################################
#Pressure for constant density using TOV#
#########################################
@jit
def PconstTOV(r,M,R):
    mem1 = np.sqrt(1.-2*M*r**2/R**3)
    mem2 = np.sqrt(1.-2*M/R)
    return 3*M/(4*math.pi*R**3)*(mem1-mem2)/(3*mem2-mem1)

############################################
#Pressure for constant density using newton#
############################################
@jit
def PconstDensNewton(r,M,R):
    return 3*M**2/(8*math.pi*R**6)*(R**2-r**2)

##########################################################
#Plot the figures regarding constant density in chapter 3#
##########################################################
def plotConstPressure():
    N = 1000
    M = 1.
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1.05)
    ax.set_xlabel('$r/R$',fontsize = 14)
    ax.set_ylabel('$P(r)/P_0$', rotation='horizontal', fontsize = 14)
    plt.ylim(-1.5,1.5)
    plt.xlim(0,1)
    R = 2.3*M
    r = np.linspace(0,R,N)
    P = PconstTOV(r,M,R)
    P0 = P[0]
    print(P0)
    plt.plot(r/R,P/P0,'g',label = '$R=2.30MG$')
    plt.plot(r/R,np.zeros(len(P)),'k--')
    R = 2.2*M
    r = np.linspace(0,R,N)
    P = PconstTOV(r,M,R)
    for i in range(1,len(P)):
        if(np.sign(P[i-1])!=np.sign(P[i])):
            index = i
    plt.plot(r[0:index]/R,P[0:index]/P0,'b',label = '$R=2.20MG$')
    plt.plot(r[index:len(P)]/R,P[index:len(P)]/P0,'b')
    R = 2.25*M
    r = np.linspace(0,R,N)
    P = PconstTOV(r,M,R)
    plt.plot(r/R,P/P0,'r',label = '$R=2.25MG$')
    
    
    plt.legend()
    
    plt.figure()
    R = 2.3*M
    r = np.linspace(0,R,N)
    P = PconstTOV(r,M,R)
    plt.xlim(0,1)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1.05)
    ax.text(0.2,0.4*10**(-1),'$R=2.3MG$',fontsize=14)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    ax.set_xlabel('$r/R$',fontsize = 14)
    ax.set_ylabel('$P(r)/P_0$', rotation='horizontal', fontsize = 14)
    plt.plot(r/R,P/P0,'r',label = 'TOV')
    P = PconstDensNewton(r,M,R)
    plt.plot(r/R,P/P0,'b',label = 'Newton')
    plt.legend()
    
    R = 10*M
    r = np.linspace(0,R,N)
    
    plt.figure()
    plt.xlim(0,1)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1.05)
    ax.text(0.2,0.4*10**(-5),'$R=10MG$',fontsize=14)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.set_xlabel('$r/R$',fontsize = 14)
    ax.set_ylabel('$P(r)/P_0$', rotation='horizontal', fontsize = 14)
    P = PconstTOV(r,M,R)
    plt.plot(r/R,P/P0,'r',label = 'TOV')
    P = PconstDensNewton(r,M,R)
    plt.plot(r/R,P/P0,'b',label = 'Newton')
    plt.legend()
    
    R = 100*M
    r = np.linspace(0,R,N)
    
    plt.figure()
    plt.xlim(0,1)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1.05)
    ax.text(0.2,0.4*10**(-9),'$R=100MG$',fontsize=14)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.set_xlabel('$r/R$',fontsize = 14)
    ax.set_ylabel('$P(r)/P_0$', rotation='horizontal', fontsize = 14)
    P = PconstTOV(r,M,R)
    plt.plot(r/R,P/P0,'r',label = 'TOV')
    P = PconstDensNewton(r,M,R)
    plt.plot(r/R,P/P0,'b',label = 'Newton')
    plt.legend()
    
    R = 1000*M
    r = np.linspace(0,R,N)
    
    plt.figure()
    plt.xlim(0,1)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1.05)
    ax.text(0.2,0.4*10**(-13),'$R=1000MG$',fontsize=14)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.set_xlabel('$r/R$',fontsize = 14)
    ax.set_ylabel('$P(r)/P_0$', rotation='horizontal', fontsize = 14)
    P = PconstTOV(r,M,R)
    plt.plot(r/R,P/P0,'r',label = 'TOV')
    P = PconstDensNewton(r,M,R)
    plt.plot(r/R,P/P0,'b',label = 'Newton')
    plt.legend()

############################################################################
#The derivative of the mass with respect to r in the non-relativistic limit#
############################################################################
@jit
def dMdrNonRel(r,P):
    if(P<0):
        return 0.
    return betaNonRel*r**2*P**(3/5)

############################################################################
#Derivative of the pressure with respect to r for the non-relativistic case#
#(TOV-equation)                                                            #
############################################################################
@jit
def dPdrTOVNonRel(r,P,M):
    if(P<0):
        return 0.
    return -alpha*P**(3/5)*r**(-2)*(1+P**(2/5)*K**(-1))*(
            (M+betaNonRel*r**3*P*K**(-1))*(1-2*R0*M*r**(-1))**(-1))

########################################################################
#Derivative of the mass with respect to r for the non-relativistic case#
#(TOV-equation)                                                        #
########################################################################
@jit
def dPdrNewtonNonRel(r,P,M):
    if(P<0):
        return 0.
    return -alpha*P**(3/5)*M*r**(-2)
    
#############################################################
#Creates the coefficients for the Runge Kutta routine in the# 
#non-relativistic case using the TOV-equation               #
#############################################################
@jit
def kTOVNonRel(r,P,M):
    kP1 = dPdrTOVNonRel(r,P,M)
    kM1 = dMdrNonRel(r,P)
    kP2 = dPdrTOVNonRel(r+h/2,P+h/2*kP1,M+h/2*kM1)
    kM2 = dMdrNonRel(r+h/2,P+h/2*kP1)
    kP3 = dPdrTOVNonRel(r+h/2,P+h/2*kP2,M+h/2*kM2)
    kM3 = dMdrNonRel(r+h/2,P+h/2*kP2)
    kP4 = dPdrTOVNonRel(r+h,P+h*kP3,M+h*kM3)
    kM4 = dMdrNonRel(r+h,P+h*kP3)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

#############################################################
#Creates the coefficients for the Runge Kutta routine in the# 
#non-relativistic case using Newton's equation              #
#############################################################
@jit
def kNewNonRel(r,P,M):
    kP1 = dPdrNewtonNonRel(r,P,M)
    kM1 = dMdrNonRel(r,P)
    kP2 = dPdrNewtonNonRel(r+h/2,P+h/2*kP1,M+h/2*kM1)
    kM2 = dMdrNonRel(r+h/2,P+h/2*kP1)
    kP3 = dPdrNewtonNonRel(r+h/2,P+h/2*kP1,M+h/2*kM2)
    kM3 = dMdrNonRel(r+h/2,P+h/2*kP2)
    kP4 = dPdrNewtonNonRel(r+h,P+h*kP3,M+h*kM3)
    kM4 = dMdrNonRel(r+h,P+h*kP3)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

###########################################################
#Function that calculates the mass and radi of a star with#
#given central pressure in the non-relativistic case      #
#Also returns a flag indicating true if                   #
#maximum iterations is reached                            #
###########################################################
@jit 
def resultsNonRel(Pc,h,nMax):
    PTOV = Pc
    PNew = Pc
    MTOV = 0.
    MNew = 0.
    rTOV = 0.
    rNew = 0.
    flagNew = False
    flagTOV = False
    for i in range(nMax):    
        MemoryMTOV = MTOV
        MemoryMNew = MNew      
        if(flagTOV == False):
            rTOV=rTOV+h
            (kPTOV1,kPTOV2,kPTOV3,kPTOV4,kMTOV1,kMTOV2,kMTOV3,kMTOV4) = kTOVNonRel(rTOV,PTOV,MTOV)
            PTOV = PTOV + h/6*(kPTOV1+2*kPTOV2+2*kPTOV3+kPTOV4)
            MTOV = MTOV + h/6*(kMTOV1+2*kMTOV2+2*kMTOV3+kMTOV4)
        
        if(flagNew == False):
            rNew = rNew+h
            (kPNew1,kPNew2,kPNew3,kPNew4,kMNew1,kMNew2,kMNew3,kMNew4) = kNewNonRel(rNew,PNew,MNew)
        
            PNew = PNew + h/6*(kPNew1+2*kPNew2+2*kPNew3+kPNew4)
            MNew = MNew + h/6*(kMNew1+2*kMNew2+2*kMNew3+kMNew4)
        
        if(np.float((np.real(PTOV))<=0 or MTOV==MemoryMTOV) and flagTOV == False):
            flagTOV = True
            
        if(np.float((np.real(PNew))<=0 or MemoryMNew == MNew) and flagNew == False):
            flagNew = True
            
        if(flagNew == True and flagTOV == True):
            break 
    if(flagNew == False):
        print("Maximum number of iterations reached")
        print("for Newton-equation with Pc = %.8f \n"%Pc)
    if(flagTOV == False):
        print("Maximum number of iterations reached")
        print("for TOV-equation with Pc = %.8f \n"%Pc)
    return rTOV,MTOV,rNew,MNew,flagTOV,flagNew
        
  

############################################################################
#Returns vectors containing the radii and mass for central pressures within#
#the range given in the non-relatiistic case                               #
############################################################################
@jit
def paramterisingNonRel(PcMin,PcMax,N,h,nMax):
    Pc = PcMin
    const = (PcMax/PcMin)**(np.float(1)/N)
    RTOV = np.zeros(N)
    RNew = np.zeros(N)
    MTOV = np.zeros(N)
    MNew = np.zeros(N)
    for i in range(N):       
        RTOV[i],MTOV[i],RNew[i],MNew[i],flagTOV,flagNew = resultsNonRel(Pc,h,nMax)
        if(flagTOV == False or flagNew == False or Pc > PcMax or Pc > PcMax):
            RTOV = RTOV[0:i]
            MTOV = MTOV[0:i]
            RNew = RNew[0:i]
            MNew = MNew[0:i]
            N = i
            break
        Pc = Pc*const
    return RTOV,MTOV,RNew,MNew,N

###########################################################
#Writes all the non-relativistic mass-radii data to a file#
###########################################################
def writeResultsToFileNonRel(PcMin,PcMax,N,h,nMax,filename):
    RTOV,MTOV,RNew,MNew,N = paramterisingNonRel(PcMin,PcMax,N,h,nMax)
    f = open(filename,'w')
    f.write(str(N)+'\n')
    for i in range(N):
        a = str(RTOV[i])
        b = str(MTOV[i])
        c = str(RNew[i])
        d = str(MNew[i])
        f.write(a+" "+b+" "+c+" "+d +"\n")
    f.close()
    return
    
########################################################################
#Reads the data from file and returns four vectors containing the      #
#mass-radius relation for both the TOV-equation and Newton's equation  #
#in the non-relativistic case                                          #
########################################################################
def readResultsFromFileNonRel(filename):
    f = open(filename,'r')
    N = int(f.readline())
    RTOV = np.zeros(N)
    MTOV = np.zeros(N)
    RNew = np.zeros(N)
    MNew = np.zeros(N)
    i = 0
    data = f.readlines()
    for line in data:
        numbers = line.split()
        RTOV[i] = numbers[0]
        MTOV[i] = numbers[1]
        RNew[i] = numbers[2]
        MNew[i] = numbers[3]
        i = i+1 
    f.close()
    return RTOV,MTOV,RNew,MNew
    
###################################          
###################################
###                             ###
###   Ultra-relativistic case   ###
###                             ###
###################################
###################################


##############################################################################
#The derivative of the mass with respect to r in the ultra-relativistic limit#
##############################################################################
@jit
def dMdrUltraRel(r,P):
    return betaUltraRel*r**2*P

############################################################################
#Derivative of the pressure with respect to r for the non-relativistic case#
#(TOV-equation)                                                            #
############################################################################
@jit
def dPdrTOVUltraRel(r,P,M):
    return -4*R0*P*r**(-2)*(M+betaUltraRel*P*r**3/3)*(1-2*R0*M*r**(-1))**(-1)

############################################################################
#Derivative of the pressure with respect to r for the non-relativistic case#
#Newton's equation                                                         #
############################################################################
@jit
def dPdrNewtonUltraRel(r,P,M):
    return -3*R0*P*M*r**(-2)

#############################################################
#Creates the coefficients for the Runge Kutta routine in the# 
#ultra-relativistic case using the TOV-equation             #
#############################################################
@jit
def kTOVUltraRel(r,P,M,h):
    kP1 = dPdrTOVUltraRel(r,P,M)
    kM1 = dMdrUltraRel(r,P)
    kP2 = dPdrTOVUltraRel(r+h/2,P+h/2*kP1,M+h/2*kM1)
    kM2 = dMdrUltraRel(r+h/2,P+h/2*kP1)
    kP3 = dPdrTOVUltraRel(r+h/2,P+h/2*kP2,M+h/2*kM2)
    kM3 = dMdrUltraRel(r+h/2,P+h/2*kP2)
    kP4 = dPdrTOVUltraRel(r+h,P+h*kP3,M+h*kM3)
    kM4 = dMdrUltraRel(r+h,P+h*kP3)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

#############################################################
#Creates the coefficients for the Runge Kutta routine in the# 
#ultra-relativistic case using the Newton's equation        #
#############################################################
@jit
def kNewUltraRel(r,P,M,h):
    kP1 = dPdrNewtonUltraRel(r,P,M)
    kM1 = dMdrUltraRel(r,P)
    kP2 = dPdrNewtonUltraRel(r+h/2,P+h/2*kP1,M+h/2*kM1)
    kM2 = dMdrUltraRel(r+h/2,P+h/2*kP1)
    kP3 = dPdrNewtonUltraRel(r+h/2,P+h/2*kP2,M+h/2*kM2)
    kM3 = dMdrUltraRel(r+h/2,P+h/2*kP2)
    kP4 = dPdrNewtonUltraRel(r+h,P+h*kP3,M+h*kM3)
    kM4 = dMdrUltraRel(r+h,P+h*kP3)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

############################################################
#Function that calculates the mass and pressure at N points#
#given a central pressure Pc as function of the radius r.  #
#Reurns 5 vectors which contains the radius and masses     #
#obtained using both Newton and TOV-equation               # 
############################################################
@jit 
def resultsUltraRel(R,R0,Pc,h):
    N = int(R/h)
    PTOV = np.zeros(N)
    PTOV[0] = Pc
    PNew = np.zeros(N)
    PNew[0] = Pc
    MTOV = np.zeros(N)
    MNew = np.zeros(N)
    
    r = np.linspace(h,R,N)
    for i in range(1,N): 
        (kPTOV1,kPTOV2,kPTOV3,kPTOV4,
        kMTOV1,kMTOV2,kMTOV3,kMTOV4) = kTOVUltraRel(r[i-1],PTOV[i-1],MTOV[i-1],h)
        
        PTOV[i] = PTOV[i-1] + h/6*(kPTOV1+2*kPTOV2+2*kPTOV3+kPTOV4)
        MTOV[i] = MTOV[i-1] + h/6*(kMTOV1+2*kMTOV2+2*kMTOV3+kMTOV4)
       
        (kPNew1,kPNew2,kPNew3,kPNew4,
        kMNew1,kMNew2,kMNew3,kMNew4) = kNewUltraRel(r[i-1],PNew[i-1],MNew[i-1],h)
        
        PNew[i] = PNew[i-1] + h/6*(kPNew1+2*kPNew2+2*kPNew3+kPNew4)
        MNew[i] = MNew[i-1] + h/6*(kMNew1+2*kMNew2+2*kMNew3+kMNew4)  
    return r,PTOV,MTOV,PNew,MNew

##############################################
#Writes all ultra-relativistic data to a file#
##############################################
def writeResultsToFileUltraRel(R,Pc,h,filename):
    r,PTOV,MTOV,PNew,MNew = resultsUltraRel(R,R0,Pc,h)
    f = open(filename,'w')
    f.write(str(len(r))+'\n')
    for i in range(len(r)):
        a = str(r[i])
        b = str(PTOV[i])
        c = str(MTOV[i])
        d = str(PNew[i])
        e = str(MNew[i])
        f.write(a+" "+b+" "+c+" "+d+" "+e+"\n")
    f.close()
    return

#######################################################################
#Reads all ultra-relativistic data from file and returns it as vectors#
#######################################################################
def readResultsFromFileUltraRel(filename):
    f = open(filename,'r')
    N = int(f.readline())
    r = np.zeros(N)
    PTOV = np.zeros(N)
    MTOV = np.zeros(N)
    PNew = np.zeros(N)
    MNew = np.zeros(N)
    i = 0
    data = f.readlines()
    for line in data:
        numbers = line.split()
        r[i] = numbers[0]
        PTOV[i] = numbers[1]
        MTOV[i] = numbers[2]
        PNew[i] = numbers[3]
        MNew[i] = numbers[4]
        i = i+1 
    f.close()
    return r,PTOV,MTOV,PNew,MNew

###########################################################################
#Calculates the analytical solution to the pressure for ultra-relativistic# 
#ideal neutron gas                                                        #
###########################################################################
@jit
def analyticPressure(r):
    const = (14*R0*betaUltraRel/3)**(-1)
    return const*r**(-2)

####################################################################
#Calculates the analytical mass solution for the ultra-relativistic# 
#ideal neutron gas                                                 #
####################################################################
@jit
def analyticMass(r):
    const = (14*R0*betaUltraRel/3)**(-1)
    return betaUltraRel*const*r

################################    
################################
###                          ###
###   Arbitrary relativity   ###
###                          ###
################################
################################

###########################################################
#The derivative of the mass M with respect to the radius r#
#for arbitrary relativity                                 #
###########################################################
@jit
def dMdr(r,epsilon):
    return beta*epsilon*r**2

###############################################################
#The derivative of the pressure P with respect to the radius r#
#for arbitrary relativity                                     #
###############################################################
@jit
def dPdrTOV(r,P,M,epsilon):
    return-R0*r**(-2)*(epsilon+P)*(M+beta*P*r**3)*(1-2*R0*M*r**(-1))**(-1)

#################################################
#The pressure as function of the fermi energy pF#
#################################################
@jit
def pressure(pF):
    return gamma*(24.*math.pi**2)**(-1)*(np.sqrt(pF**2+1)*(2*pF**3-3*pF)+3*np.log(pF+np.sqrt(pF**2+1)))

###################################################################
#Function that creates the coefficients in the Runge Kutta routine# 
###################################################################
@jit
def k(r,P,M,epsilon,h):
    kP1 = dPdrTOV(r,P,M,epsilon)
    kM1 = dMdr(r,epsilon)
    kP2 = dPdrTOV(r+h/2,P+h/2*kP1,M+h/2*kM1,epsilon)
    kM2 = dMdr(r+h/2,epsilon)
    kP3 = dPdrTOV(r+h/2,P+h/2*kP2,M+h/2*kM2,epsilon)
    kM3 = dMdr(r+h/2,epsilon)
    kP4 = dPdrTOV(r+h/2,P+h*kP3,M+h*kM3,epsilon)
    kM4 = dMdr(r+h,epsilon)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

#######################################################
#The energy density as a function of fermi momentum pF#
#######################################################
@jit
def energyDensity(pF):
    return gamma*(24.*math.pi**2)**(-1)*(np.sqrt(pF**2+1)*(6*pF**3+3*pF)-3*np.log(pF+np.sqrt(pF**2+1)))

def f(P):
    return lambda pF: pressure(pF)-P

#############################################################
#Function that returns 2 vectors containing the mass-radius #
#relation and a flag indicating if the maximum number of    #
#iterations is reached                                      # 
#############################################################
def results(Pc,h,nMax,pFMax):
    P = Pc
    M = 0.
    r = 0.
    pF = pFMax
    flag=True
    for i in range(nMax):
        Memoryr = r
        MemoryM = M
        r = r+h
        if(np.sign(f(P)(0))!=np.sign(f(P)(pFMax))):
            pF = op.brentq(f(P),0,pFMax)
        else:
            pF=0
        epsilon = energyDensity(pF)
        kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4 = k(r,P,M,epsilon,h)
        P = P+h/6*(kP1+2*kP2+2*kP3+kP4)
        M = M+h/6*(kM1+2*kM2+2*kM3+kM4)
        if(np.float(np.real(P))<=0 or M==MemoryM):
            r = Memoryr
            flag = False
            break
    if(flag == True):
            print("Maximum number of iterations reached")
            print("for TOV-equation with Pc = %.8f"%Pc)
            print("with arbitrary relativity \n")
    return r, np.float(np.real(M)),flag
  
        
###############################################################
#Does the same as parametrisingNonRel for arbitrary relativity#
###############################################################
@jit    
def paramterising(PcMin,PcMax,N,h,nMax,pFMax):
    Pc = PcMin
    const = (np.float(PcMax)/PcMin)**(np.float(1)/N)
    R = np.zeros(N)
    M = np.zeros(N)
    for i in range(N):
        R[i],M[i],flag = results(Pc,h,nMax,pFMax)
        if(flag==True):
            print(R[i])
            R = R[0:i]
            M = M[0:i]
            N = i
            break
        Pc = Pc*const
    return R,M,N

#####################
#Writes data to file#
#####################
def writeResultsToFile(PcMin,PcMax,N,h,nMax,pFMax,filename):
    R,M,N = paramterising(PcMin,PcMax,N,h,nMax,pFMax)
    f = open(filename,'w')
    f.write(str(N)+'\n')
    for i in range(N):
        a = str(R[i])
        b = str(M[i])
        f.write(a+" "+b+"\n")
    f.close()
    return 
 
######################
#Reads data from file#
######################  
def readResultsFromFile(filename):
    f = open(filename,'r')
    N = int(f.readline())
    R= np.zeros(N)
    M = np.zeros(N)
    i = 0
    data = f.readlines()
    for line in data:
        numbers = line.split()
        R[i] = numbers[0]
        M[i] = numbers[1]
        i = i+1 
    f.close()
    return R,M

####################################################################
#Reads data from file and creates plots of the mass radius relation# 
#for arbitrary relativity and the non-relativistic case            #
####################################################################
def createPlots(filename,filenameNonRel,filenameUltraRel,PcMax,h,nMax,pFMax):
    plotConstPressure()
    plotdPde(pFMax,N)
    plt.figure()
    plt.ylim(0,1.5)
    RTOVnonRel,MTOVnonRel,RNewNonRel,MNewNonRel = readResultsFromFileNonRel(filenameNonRel)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$R$[km]', fontsize = 14)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$M/M_\odot$', rotation='horizontal',fontsize = 14)
    
    plt.plot(RTOVnonRel,MTOVnonRel,'r',label='TOV non-relativistic')
    a,b,c,d,e,f=resultsNonRel(0.0113,h,nMax)
    plt.scatter(a,b,marker='x',color='green')
    plt.plot(RNewNonRel,MNewNonRel,'b',label = 'Newton non-relativistic')
    R,M = readResultsFromFile(filename)
    for i in range(len(M)):
        if(M[i]==0.0):
            M[i] = None
    plt.plot(R,M,'y',label='Arbitrary relativity')
    MMax = max(M)
    RMax = R[np.argmax(M)]
    print(RMax,np.argmax(M))
    print("The biggest possible mass is (arbitrary rel) %.3f sun masses with radi %.3fkm"%(MMax,RMax))
    plt.legend()
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.0)
    ax.set_xlabel('$r$[km]', fontsize = 14)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(r)$', rotation='horizontal',fontsize = 14)
    RultraRel,PultraRelTOV,MultraRelTOV,PultraRelNew,MultraRelNew = readResultsFromFileUltraRel(filenameUltraRel)
    PultraRelAnal = analyticPressure(RultraRel)
    MultraRelAnal = analyticMass(RultraRel)
    plt.ylim(0,max(PultraRelTOV)+max(PultraRelTOV)*0.1)
    plt.xlim(0,max(RultraRel))
    plt.plot(RultraRel,PultraRelTOV,'r',label = 'TOV')
    plt.plot(RultraRel,PultraRelNew,'b',label = 'Newton')
    plt.plot(RultraRel,PultraRelAnal,'g',label = 'Analytic')
    plt.legend()
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.08, -0.0)
    ax.set_xlabel('$r$ [km]', fontsize = 14)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    plt.ylim(0,8)
    plt.xlim(0,15)
    ax.set_ylabel('$M/M_\odot$', rotation='horizontal',fontsize = 14)
    plt.plot(RultraRel,MultraRelTOV,'r',label = 'TOV')
    plt.plot(RultraRel,MultraRelNew,'b',label = 'Newton')
    plt.plot(RultraRel,MultraRelAnal,'g',label='Analytic')
    plt.legend()
    return                          

@jit
def dPde(pFMax,N):
    pF = np.linspace(0,pFMax,N+2)
    dPde = np.ones(N+2)
    for i in range(1,N+1):
        dP = pressure(pF[i+1])-pressure(pF[i-1])
        de = energyDensity(pF[i+1])-energyDensity(pF[i-1])
        dPde[i] = dP/de
    return pF[1:N+1],dPde[1:N+1]


@jit
def plotdPde(pFMax,N):
    pF,Pde = dPde(pFMax,N)
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$\\bar{p}_F$',fontsize = 14)
    ay = plt.gca()
    ay.yaxis.set_label_coords(-0.05, 1.0)
    ay.set_ylabel('d$\\bar{P}/$d$\\bar{\epsilon}$', rotation='horizontal', fontsize = 14)
    plt.plot(pF,Pde)

###############################
###############################
###                         ###
###   Program starts here   ###
###                         ###
###############################
###############################



alpha = 1.                  #alpha, beta and gamma is given 
beta = 1.1426               #by calculations in thesis
betaNonRel = 0.7779         
betaUltraRel = 1.
gamma = 1.

Pc = 0.1                    #Central pressure used in ultra relativistic star   
R = 30                      #Radius we integrate up to for ultra relativistic star                                   
R0 = 1.477                  #MG/c^2
K = 0.67704                 #Reffered to as K_NR "bar" in thesis
h = 0.0001                  #Step length in the Runge Cutta solver
nMax = int(100/h)           #Maximum number of iterations in solver
PcMin = 1.*10**(-6)         #Smallest central pressure we solve for
PcMax = 1.*10**(4)          #Largest central pressure we solve for
N = 300                     #Number of data points in plot                       
pFMax = 10.                 #Maximum fermimomentum used in root finding function

filenameNonRel = "massRadiRelationNonRel.txt"
filenameUltraRel = "massRadiRelationUltraRel.txt"
filename = "massRadiRelation.txt"

writeResultsToFileNonRel(PcMin,PcMax,N,h,nMax,filenameNonRel)
writeResultsToFileUltraRel(R,Pc,h,filenameUltraRel)
writeResultsToFile(PcMin,PcMax,N,h,nMax,pFMax,filename) 
createPlots(filename,filenameNonRel,filenameUltraRel,PcMax,h,nMax,pFMax)
plt.show()

print("\nTime spent:")
print(time.clock()-t0)