# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import time
from numba import jit
from scipy import optimize as op
import math
from scipy import interpolate 

##########################################################
##########################################################
###                                                    ###
###   Just press play to create all plots in thesis!   ###
###                                                    ###
##########################################################
##########################################################

#variable used to time the program
t0 = time.clock()

##############################
###                        ###
###   Defining functions   ###
###                        ###
##############################

#######################################################
#Function that takes inn the hyperon fermi momenta and# 
#sets the ones that are negative to zero              #
#######################################################
def isNegative(kFhyperonVec):
    klambda,ksminus,ks0,ksplus,kximinus,kxi0 = kFhyperonVec[:]
    if(ksminus<0):
        ksminus = 0
    if(klambda<0):
        klambda = 0
    if(ks0<0):
        ks0 = 0
    if(ksplus<0):
        ksplus = 0 
    if(kximinus<0):
        kximinus = 0
    if(kxi0<0):
        kxi0 = 0 
    return [klambda,ksminus,ks0,ksplus,kximinus,kxi0]

###############################################################
#Takes in a number in fm^exponent the exponent and returns the# 
#value in MeV^-exponent                                       #
###############################################################
def fmToMeV(number,exponent):
    if(exponent>0):
        return (number**(1./exponent)/197.33)**exponent
    else:
        return (number**(-1./exponent)*197.33)**(-exponent)

#############################
#Inverse function of fmToMeV#
#############################
def MeVtoFm(number,exponent):
    if(exponent>0):
        return (number**(1./exponent)/197.33)**exponent
    else:
        return (number**(-1./exponent)*197.33)**(-exponent)

################################################
#Takes in Fermi momenta and returns the density#
################################################
def density(kF):
    return kF**3/(3*math.pi**2)

###########################################################
#Takes in a particle density and returns the Fermi momenta#
###########################################################
def momentaFromDensity(dens):
    return (3*math.pi**2*dens)**(1./3)
    
##############################################################
#Takes in the Fermi momenta and effective mass for a particle# 
#and returns its contribution to the scalar density          #
##############################################################
def scalarDensityFunc(kF,mStar):
    if(kF == 0 or mStar == 0):
        return 0.
    else:
        mem1 = mStar/(2*math.pi**2)
        mem2 = np.sqrt(kF**2+mStar**2)*kF
        mem3 = mStar**2*np.arcsinh(kF/mStar)
        return mem1*(mem2-mem3)  

##################################
#Returns the total scalar density#
##################################
@jit
def scalarDensity(omega,rho,dens,ke,kMu,kn,kp,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp):
    mStarLambda = mLambda-gSigma*xSigma[0]*sigmaField(mStar,gSigma)
    mStarS = mS-gSigma*xSigma[1]*sigmaField(mStar,gSigma)
    mStarXi = mXi-gSigma*xSigma[2]*sigmaField(mStar,gSigma)
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    mem1 = scalarDensityFunc(kn,mStar)+scalarDensityFunc(kp,mStar)
    mem2 = xSigma[0]*scalarDensityFunc(kFhyperonVec[0],mStarLambda)
    mem3 = xSigma[1]*(scalarDensityFunc(kFhyperonVec[1],mStarS)+scalarDensityFunc(kFhyperonVec[2],mStarS)+scalarDensityFunc(kFhyperonVec[3],mStarS))
    mem4 = xSigma[2]*(scalarDensityFunc(kFhyperonVec[4],mStarXi)+scalarDensityFunc(kFhyperonVec[5],mStarXi))
    return mem1+mem2+mem3+mem4

#############################################
#Calculates the Fermi momenta for the proton#
#############################################
@jit
def kpFunc(ke,kMu,kFhyperonVec):
    if(ke<0):
        ke=0
    mem1 = kFhyperonVec[1]**3-kFhyperonVec[3]**3+kFhyperonVec[4]**3
    if(ke**3+kMu**3+mem1<0):
        return 0
    return (ke**3+kMu**3+mem1)**(1./3)

#################################################
#Gives the expectation value for the sigma field#
#################################################
@jit
def sigmaField(mStar,gSigma):
    return (m-mStar)/gSigma
  
###############################################################
#Gives a function for the expectation of the omega field to be# 
#used by root solver function.                                #
###############################################################
@jit  
def omegaFieldFunc(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xOmega,hyp):
    kMu = kMuFunc(ke)
    if(ke<0):
        ke=0
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    kFhyperonVec = isNegative(kFhyperonVec)
    kp = kpFunc(ke,kMu,kFhyperonVec)
    mem = 0.
    mem+=density(kn)+density(kp)
    mem+=xOmega[0]*density(kFhyperonVec[0])
    mem+=xOmega[1]*(density(kFhyperonVec[1])+density(kFhyperonVec[2])+density(kFhyperonVec[3]))
    mem+=xOmega[2]*(density(kFhyperonVec[4])+density(kFhyperonVec[5]))
    return (omega-mem*gOmega*mOmega**(-2.))

##########################################
#Returns the expectation of the rho field#
##########################################
@jit
def rhoField(kn,kp,ksminus,ksplus,kximinus,kxi0,gRho,xRho):
    return 0.5*gRho*mRho**(-2.)*(density(kn)-density(kp)+2*xRho[1]*density(ksminus)-2*xRho[1]*density(ksplus)+xRho[2]*density(kximinus)-xRho[2]*density(kxi0))
 
##########################################################
#Gives function for the expectation of the rho field that# 
#can be used by root solver                              #
##########################################################
@jit
def rhoFieldFunc(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp):
    if(ke<0):
        ke=0
    kMu = kMuFunc(ke)
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    kFhyperonVec = isNegative(kFhyperonVec)
    kp = kpFunc(ke,kMu,kFhyperonVec)
    mem1 = density(kn)-density(kp)
    mem2 = 2*xRho[1]*density(kFhyperonVec[1])-2*xRho[1]*density(kFhyperonVec[3])
    mem3 = xRho[2]*density(kFhyperonVec[4])-xRho[2]*density(kFhyperonVec[5])
    return (rho-0.5*gRho*mRho**(-2.)*(mem1+mem2+mem3))

############################################
#Returns the Fermi momenta for the electron#
############################################
@jit
def keFermi(dens,kFvec):
    kF3 = dens*3*math.pi**2
    mem1 = kFvec[1]**3+kFvec[2]**3
    mem2 = kFvec[4]**3+2*kFvec[5]**3
    mem3 = kFvec[6]**3+2*kFvec[8]**3
    mem4 = kFvec[9]**3
    return (kF3-mem1-mem2-mem3-mem4)**(1./3)

####################################################
#Gives function for the electron Fermi momenta that# 
#can be used by root solver                        #
####################################################
@jit
def keFunc(dens,omega,rho,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp):
    if(ke<0):
        ke=0
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    kFhyperonVec = isNegative(kFhyperonVec)
    kMu = kMuFunc(ke)
    kF3 = dens*3*math.pi**2
    mem1 = kMu**3+kn**3
    mem2 = kFhyperonVec[0]**3
    mem3 = 2*kFhyperonVec[1]**3+kFhyperonVec[2]**3
    mem4 = 2*kFhyperonVec[4]**3+kFhyperonVec[5]**3
    if(kF3-mem1-mem2-mem3-mem4<0):
        return ke
    return (ke-(kF3-mem1-mem2-mem3-mem4)**(1./3))

########################################
#Returns the Fermi momenta for the muon#
########################################
@jit
def kMuFunc(ke):
    if(ke**2+me**2-mMu**2>0):
        return np.sqrt(ke**2+me**2-mMu**2)
    else:
        return 0.

###################################################
#Gives function for the neutron Fermi momenta that# 
#can be used by root solver                       #
###################################################
@jit
def knFunc(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xRho,hyp):
    if(ke<0):
        ke=0
    kMu = kMuFunc(ke)
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    kFhyperonVec = isNegative(kFhyperonVec)
    kp = kpFunc(ke,kMu,kFhyperonVec)
    mem = -gRho*rho+np.sqrt(ke**2+me**2)+np.sqrt(kp**2+mStar**2)
    if(mem**2-mStar**2<=0):
        return kn
    return (kn-(mem**2-mStar**2)**(1./2))

#############################################################
#Gives function for the effective mass for the nucleons that# 
#can be used by root solver                                 #
#############################################################
@jit
def mStarFunc(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,b,c,xSigma,xOmega,xRho,hyp):
    if(ke<0):
        ke=0
    kMu = kMuFunc(ke)
    mem1 = (m-mStar)
    mem2 = (gSigma**2*mSigma**(-2))
    kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    kFhyperonVec = isNegative(kFhyperonVec)
    kp = kpFunc(ke,kMu,kFhyperonVec)
    mem3 = scalarDensity(omega,rho,dens,ke,kMu,kn,kp,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp)
    mem4 = b*m*mem1**2+c*mem1**3
    if(mem3+mem4==0):
        return mem1
    return (mem1-(mem2*(mem3+mem4)))
  
################################################
#Returns the chemical potential for the neutron# 
################################################
@jit
def munFunc(omega,rho,kn,mStar,gOmega,gRho,xOmega,xRho):
    return np.sqrt(kn**2+mStar**2)+gOmega*omega+0.5*gRho*rho

###############################################################################
#Sets the functions knFunc, mStarFunc, rhoFieldFunc, keFunc and omegaFieldFunc#
#together to one vector function to be used in root solver function           #                                             
###############################################################################
def knMstarFunc(keOrDens,gSigma,gOmega,gRho,xOmega,xRho,b,c,nr,hyp):
    if(hyp==True and nr=='n'):
        return lambda var:[knFunc(var[0],var[1],keOrDens,var[2],var[3],var[4],gOmega,gSigma,gRho,xRho,hyp),
                            mStarFunc(var[0],var[1],keOrDens,var[2],var[3],var[4],gSigma,gOmega,gRho,b,c,xSigma,xOmega,xRho,hyp),
                            rhoFieldFunc(var[0],var[1],keOrDens,var[2],var[3],var[4],gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp),
                            keFunc(keOrDens,var[0],var[1],var[2],var[3],var[4],gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp),
                            omegaFieldFunc(var[0],var[1],keOrDens,var[2],var[3],var[4],gSigma,gOmega,gRho,xOmega,hyp)]
        
    elif(nr=='n'):
        return lambda var:[knFunc(var[0],var[1],0,keOrDens,var[2],var[3],gOmega,gSigma,gRho,xRho,hyp),
                            mStarFunc(var[0],var[1],0,keOrDens,var[2],var[3],gSigma,gOmega,gRho,b,c,xSigma,xOmega,xRho,hyp),
                            rhoFieldFunc(var[0],var[1],0,keOrDens,var[2],var[3],gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp),
                            omegaFieldFunc(var[0],var[1],0,keOrDens,var[2],var[3],gSigma,gOmega,gRho,xOmega,hyp)]
    elif(nr=='r'):
        kMu = kMuFunc(keOrDens)
        kp = kpFunc(keOrDens,kMu,[0,0,0,0,0,0])
        return lambda var:[knFunc(var[0],var[1],0,keOrDens,var[2],var[3],gOmega,gSigma,gRho,xRho,hyp),
                            renormalizedMstarFunc(keOrDens,var[2],kp,var[3],gSigma,gOmega,gRho,b,c),
                            rhoFieldFunc(var[0],var[1],0,keOrDens,var[2],var[3],gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp),
                            omegaFieldFunc(var[0],var[1],0,keOrDens,var[2],var[3],gSigma,gOmega,gRho,xOmega,hyp)]

#################################################
#Returns the Fermi momenta for a type of hyperon#
#################################################
@jit
def fermiHyperonFunc(omega,rho,ke,kn,q,I,mStar,mStarB,gOmega,gRho,xSigma,xOmega,xRho,index):
    mun = munFunc(omega,rho,kn,mStar,gOmega,gRho,xOmega,xRho)
    mem = (mun-q*np.sqrt(ke**2+me**2)-gOmega*xOmega[index]*omega-gRho*xRho[index]*I*rho)**2-mStarB**2
    if(mem>0):
        return np.sqrt(mem)
    else:
        return 0
   
########################################################
#Returns the Fermi momenta for the hyperons in a vector# 
########################################################
def fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp):
    if(hyp==False):
        return [0,0,0,0,0,0]
    mStarLambda = mLambda-gSigma*xSigma[0]*sigmaField(mStar,gSigma)
    mStarS = mS-gSigma*xSigma[1]*sigmaField(mStar,gSigma)
    mStarXi = mXi-gSigma*xSigma[2]*sigmaField(mStar,gSigma)
    klambda = fermiHyperonFunc(omega,rho,ke,kn,0,0,mStar,mStarLambda,gOmega,gRho,xSigma,xOmega,xRho,0)
    ksminus = fermiHyperonFunc(omega,rho,ke,kn,-1,1,mStar,mStarS,gOmega,gRho,xSigma,xOmega,xRho,1)
    ks0 = fermiHyperonFunc(omega,rho,ke,kn,0,0,mStar,mStarS,gOmega,gRho,xSigma,xOmega,xRho,1)
    ksplus = fermiHyperonFunc(omega,rho,ke,kn,1,-1,mStar,mStarS,gOmega,gRho,xSigma,xOmega,xRho,1)
    kximinus = fermiHyperonFunc(omega,rho,ke,kn,-1,0.5,mStar,mStarXi,gOmega,gRho,xSigma,xOmega,xRho,2)
    kxi0 = fermiHyperonFunc(omega,rho,ke,kn,0,-0.5,mStar,mStarXi,gOmega,gRho,xSigma,xOmega,xRho,2)    
    return [klambda,ksminus,ks0,ksplus,kximinus,kxi0]
    
#########################
#Returns the omega field#
#########################
@jit
def omegaField(kFvec,gOmega,xOmega):
    omega = 0
    omega+=density(kFvec[2])+density(kFvec[3])
    omega+=xOmega[0]*density(kFvec[4])
    omega+=xOmega[1]*(density(kFvec[5])+density(kFvec[6])+density(kFvec[7]))
    omega+=xOmega[2]*(density(kFvec[8])+density(kFvec[9]))
    return gOmega*mOmega**(-2.)*omega
  
################################################
#Returns all Fermi momenta and effective masses#
################################################
@jit
def fermiMomentaAndEffectiveMass(keOrDens,guesses,gSigma,gOmega,gRho,xSigma,xOmega,xRho,b,c,nr,hyp):
    if(hyp==False):
        ke = keOrDens
        xSigma = [0,0,0]
        xRho = [0,0,0]
        xOmega = [0,0,0]
        solution = op.root(knMstarFunc(ke,gSigma,gOmega,gRho,xOmega,xRho,b,c,nr,hyp),guesses)
        omega,rho,kn,mStar = solution.x
        kFhyperonVec = [0,0,0,0,0,0] 
        
        
    else:
        dens = keOrDens
        solution = op.root(knMstarFunc(dens,gSigma,gOmega,gRho,xOmega,xRho,b,c,nr,hyp),guesses,options={'factor':0.00003})
        omega,rho,ke,kn,mStar = solution.x
        kFhyperonVec = fermiHyperon(omega,rho,dens,ke,kn,mStar,gSigma,gOmega,gRho,xSigma,xOmega,xRho,hyp) 
    kMu = kMuFunc(ke)       
    klambda,ksminus,ks0,ksplus,kximinus,kxi0 = kFhyperonVec[:]
    kp = kpFunc(ke,kMu,kFhyperonVec)
    mStarLambda = mLambda-gSigma*xSigma[0]*sigmaField(mStar,gSigma)
    mStarS = mS-gSigma*xSigma[1]*sigmaField(mStar,gSigma)
    mStarXi = mXi-gSigma*xSigma[2]*sigmaField(mStar,gSigma)

    return omega,rho,ke,kMu,kn,kp,klambda,ksminus,ks0,ksplus,kximinus,kxi0,mStar,mStarLambda,mStarS,mStarXi,solution.success

################################################
#Returns the Fermi momenta for a free Fermi gas#
################################################
@jit
def freeFermiGasEnergyDensity(kF,mStar):
    if(kF==0):
        return 0.
    else:
        mem1 = (24*math.pi**2)**(-1)
        mem2 = np.sqrt(kF**2+mStar**2)
        mem3 = 6*kF**3+3*mStar**2*kF
        if(mStar==0):
            mem4 = 0.
        else:
            mem4 = 3*mStar**4*np.arcsinh(kF/mStar)
        return mem1*(mem2*mem3-mem4)

###########################################
#Returns the pressure for a Free fermi gas#
###########################################
@jit
def freeFermiGasPressure(kF,mStar):
    if(kF==0):
        return 0.
    else:
        mem1 = (24*math.pi**2)**(-1)
        mem2 = np.sqrt(kF**2+mStar**2)
        mem3 = 2*kF**3-3*mStar**2*kF
        if(mStar==0):
            mem4 = 0.
        else:
            mem4 = 3*mStar**4*np.arcsinh(kF/mStar)
        return mem1*(mem2*mem3+mem4)

######################
#Returns the pressure#
######################
@jit
def pressure(kFvec,mStar,mStarLambda,mStarS,mStarXi,gSigma,gOmega,gRho,b,c,xRho,nr,hyp):
    if(kFvec[0]==0 and kFvec[3]==0):
        return 0.
    elif(nr=='r'):
        return renormalizedPressure(kFvec,mStar,gSigma,gOmega,gRho,b,c)
    else:
        ke,kMu,kn,kp,klambda,ksminus,ks0,ksplus,kximinus,kxi0 = kFvec[:]
        sigma = sigmaField(mStar,gSigma)
        omega = omegaField(kFvec,gOmega,xOmega)
        rho = rhoField(kn,kp,ksminus,ksplus,kximinus,kxi0,gRho,xRho)
        mem1 = -0.5*mSigma**2*sigma**2+1./3*b*m*(gSigma*sigma)**3+1./4*c*(gSigma*sigma)**4
        mem2 = 0.5*mOmega**2*omega**2+0.5*mRho**2*rho**2
        mem3 = freeFermiGasPressure(ke,me)
        mem4 = freeFermiGasPressure(kMu,mMu)
        mem5 = freeFermiGasPressure(kn,mStar)
        mem6 = freeFermiGasPressure(kp,mStar)
        mem7 = freeFermiGasPressure(klambda,mStarLambda)
        mem8 = freeFermiGasPressure(ksminus,mStarS)
        mem9 = freeFermiGasPressure(ks0,mStarS)
        mem10 = freeFermiGasPressure(ksplus,mStarS)
        mem11 = freeFermiGasPressure(kximinus,mStarXi)
        mem12 = freeFermiGasPressure(kxi0,mStarXi)
        return mem1+mem2+mem3+mem4+mem5+mem6+mem7+mem8+mem9+mem10+mem11+mem12

############################
#Returns the energy density#
############################
@jit
def energyDensity(kFvec,mStar,mStarLambda,mStarS,mStarXi,gSigma,gOmega,gRho,b,c,xRho,nr,hyp):
    if(kFvec[0]==0 and kFvec[3]==0):
        return 0.
    elif(nr=='r'):
        return renormalizedEnergyDensity(kFvec,mStar,gSigma,gOmega,gRho,b,c)
    else:
        ke,kMu,kn,kp,klambda,ksminus,ks0,ksplus,kximinus,kxi0 = kFvec[:]
        sigma = sigmaField(mStar,gSigma)
        omega = omegaField(kFvec,gOmega,xOmega)
        rho = rhoField(kn,kp,ksminus,ksplus,kximinus,kxi0,gRho,xRho)
        mem1 = 0.5*mSigma**2*sigma**2-1./3*b*m*(gSigma*sigma)**3-1./4*c*(gSigma*sigma)**4
        mem2 = 0.5*mOmega**2*omega**2+0.5*mRho**2*rho**2
        mem3 = freeFermiGasEnergyDensity(ke,me)
        mem4 = freeFermiGasEnergyDensity(kMu,mMu)
        mem5 = freeFermiGasEnergyDensity(kn,mStar)
        mem6 = freeFermiGasEnergyDensity(kp,mStar)
        mem7 = freeFermiGasEnergyDensity(klambda,mStarLambda)
        mem8 = freeFermiGasEnergyDensity(ksminus,mStarS)
        mem9 = freeFermiGasEnergyDensity(ks0,mStarS)
        mem10 = freeFermiGasEnergyDensity(ksplus,mStarS)
        mem11 = freeFermiGasEnergyDensity(kximinus,mStarXi)
        mem12 = freeFermiGasEnergyDensity(kxi0,mStarXi)
        return mem1+mem2+mem3+mem4+mem5+mem6+mem7+mem8+mem9+mem10+mem11+mem12

###########################
###                     ###
###   Renromalization   ###
###                     ###
###########################

####################
#f'' from chapter 7#
####################
@jit
def fDoubleDer(mStar,gSigma,b,c):
    return 2*b*m*gSigma**3*sigmaField(mStar,gSigma)+3*c*gSigma**4*sigmaField(mStar,gSigma)**2

#####################
#f''' from chapter 7#
#####################
@jit
def fTrippelDer(mStar,gSigma,b,c):
    return 2*b*gSigma**3+6*c*gSigma**4*sigmaField(mStar,gSigma)

##################
#U from chapter 7#
##################
@jit    
def U(mStar,gSigma):
    if(sigmaField(mStar,gSigma)<=0.):
        return 0.
    mem1 = -(2*math.pi)**(-2)
    if(mStar==0.):
        mem2 = 0.
    else:
        mem2 = mStar**4*np.log(mStar/m)
    mem3 = m**3*(m-mStar)-7./2*m**2*(m-mStar)**2
    mem4 = 13./3*m*(m-mStar)**3-25./12*(m-mStar)**4
    return mem1*(mem2+mem3+mem4)

####################################
#The derivative of U from chapter 7#
####################################
@jit
def derU(mStar):
    mem1 = (math.pi)**(-2)
    if(mStar <= 0.):
        mem2 = 0.
    else:
        mem2 = mStar**3*np.log(mStar/m)
    mem3 = m**2*(m-mStar)
    mem4 = -5./2*m*(m-mStar)**2
    mem5 = 11./6*(m-mStar)**3
    return mem1*(mem2+mem3+mem4+mem5) 
    


##################
#V from chapter 7#
##################
@jit
def V(mStar,gSigma,b,c):
    mem1 = fDoubleDer(mStar,gSigma,b,c)
    if(mSigma**2-mem1<=0):
        mem2 = 0.
    else:
        mSigmaStar = np.sqrt(mSigma**2-mem1)
        mem2 = mSigmaStar**4*np.log(mSigmaStar/mSigma)
    mem3 = 2*b*m*gSigma**3*sigmaField(mStar,gSigma)
    mem4 = 3*c*gSigma**4*sigmaField(mStar,gSigma)**2
    return (mem2-0.5*(-mSigma**2*mem1+3./2*mem1**2-(mem3**3+3*mem3**2*mem4)/(3*mSigma**2)-mem3**4/(12*mSigma**4)))/(32*math.pi**2)

################################
#Derivative of V from chapter 7#
################################
@jit
def derV(mStar,gSigma,b,c):
    mem1 = fDoubleDer(mStar,gSigma,b,c)
    if(mSigma**2-mem1<0):
        mSigmaStar = 0.
    else:
        mSigmaStar = np.sqrt(mSigma**2-mem1)
    mem2 = fTrippelDer(mStar,gSigma,b,c)
    mem3 = 2*b*m*gSigma**3
    mem4 = 3*c*gSigma**4
    mem5 = sigmaField(mStar,gSigma)
    if(mSigmaStar <= 0.):
        mem6 = 0.
    else:
        mem6 = (4*mSigmaStar**3*np.log(mSigmaStar/mSigma)+mSigmaStar**3)*mem2/(2*gSigma*mSigmaStar)
    return -(mem6+0.5/gSigma*(-mSigma**2*mem2+3*mem2*mem1-(mem3**3*mem5**2+4*mem3**2*mem4*mem5**3)/mSigma**2-(mem3**4*mem5**3)/(3*mSigma**4)))/(32*math.pi**2)

################################################################
#Same as mStarFunc except that renormalization is accounted for#
################################################################
@jit
def renormalizedMstarFunc(ke,kn,kp,mStar,gSigma,gOmega,gRho,b,c):
    kMu = kMuFunc(ke)
    mem1 = m-mStar
    mem2 = gSigma**2*mSigma**(-2)
    mem3 = b*m*mem1**2
    mem4 = c*mem1**3
    omega = omegaField([0,0,kn,kp,0,0,0,0,0,0],gOmega,xOmega)
    rho = rhoField(kn,kp,0,0,0,0,gRho,xRho)
    mem5 = scalarDensity(omega,rho,0,ke,kMu,kn,kp,mStar,gSigma,gOmega,gRho,[0,0,0],[0,0,0],[0,0,0],False)
    return 1.-mem1/(mem2*(mem3+mem4+mem5-derU(mStar)-derV(mStar,gSigma,b,c)))

####################################################################
#Same as energyDensity except that renormalization is accounted for#
####################################################################
@jit
def renormalizedEnergyDensity(kFvec,mStar,gSigma,gOmega,gRho,b,c):
    mem1 = energyDensity(kFvec,mStar,0,0,0,gSigma,gOmega,gRho,b,c,[0,0,0],'n',False)
    mem2 = V(mStar,gSigma,b,c)
    mem4 = U(mStar,gSigma)
    return mem1+mem2+mem4

###############################################################
#Same as pressure except that renormalization is accounted for#
###############################################################
@jit
def renormalizedPressure(kFvec,mStar,gSigma,gOmega,gRho,b,c):
    mem1 = pressure(kFvec,mStar,0,0,0,gSigma,gOmega,gRho,b,c,[0,0,0],'n',False)
    mem2 = V(mStar,gSigma,b,c)
    mem4 = U(mStar,gSigma)
    return mem1-mem2-mem4
    



#############################################################
###                                                       ###
###   Functions used to calculate the coupling constants  ###
###                                                       ###
#############################################################

################################################################
#Same as mStarFunc except that it only holds for nuclear matter#
################################################################
@jit
def nuclearMatterMassFunc(kF,mStar,gSigma,gOmega,gRho,b,c,nr,hyp):
    mem1 = (m-mStar)
    mem2 = gSigma**2*mSigma**(-2)
    omega = omegaField([0,0,kF,kF,0,0,0,0,0,0],gOmega,xOmega)
    rho = rhoField(kF,kF,0,0,0,0,gRho,[0,0,0])
    mem3 = scalarDensity(omega,rho,0,0,0,kF,kF,mStar,gSigma,gOmega,gRho,[0,0,0],[0,0,0],[0,0,0],hyp)
    mem4 = b*m*mem1**2+c*mem1**3    
    if(nr=='r'):
        return -mem1+mem2*(mem3+mem4-derU(mStar)-derV(mStar,gSigma,b,c))
    return -mem1+mem2*(mem3+mem4)

#####################################################
#Finds the effective nucleon mass for nuclear matter#
#####################################################
def nuclearEffectiveMass(kF,gSigma,gOmega,gRho,b,c,nr,hyp):
    if(kF==0):
        return 1.
    else:
        func = lambda mStar: nuclearMatterMassFunc(kF,mStar,gSigma,gOmega,gRho,b,c,nr,hyp)
        if(np.sign(func(0))==np.sign(func(1))):
            return 0.
        else:
            dsad = op.brentq(func,0,1)
            return dsad

############################################################################
#Returns the effective mass, energy density and pressure for nuclear matter#
#for a given Fermi momenta                                                 #
############################################################################
@jit
def nuclearMatterEoS(kF,gSigma,gOmega,gRho,b,c,nr,hyp):
    mStar = nuclearEffectiveMass(kF,gSigma,gOmega,gRho,b,c,nr,hyp)
    epsilon = energyDensity([0,0,kF,kF,0,0,0,0,0,0],mStar,0,0,0,gSigma,gOmega,gRho,b,c,[0,0,0],nr,hyp)
    P = pressure([0,0,kF,kF,0,0,0,0,0,0],mStar,0,0,0,gSigma,gOmega,gRho,b,c,[0,0,0],nr,hyp)
    return mStar,epsilon,P

###############################################################################
#Finds the rho coupling given a Fermi momenta and effective mass at saturation#
#as well as the symmetry energy coefficient as                                #
###############################################################################
@jit
def findgRho(kFsat,mStar,aS):
    mem1 = aS-kFsat**2/(6*np.sqrt(kFsat**2+mStar**2))
    mem2 = 12*math.pi**2*kFsat**(-3)
    gRho = mRho*np.sqrt(mem1*mem2)
    return gRho

#############################################################
#Function used to calculate the effective mass at saturation#
#############################################################
@jit
def mStarFuncCouplings(kFsat,gSigma,gOmega,gRho,b,c,nr):
    return nuclearMatterEoS(kFsat,gSigma,gOmega,gRho,b,c,nr,False)[0]

##########################################
#Returns the binding energy at saturation#
##########################################
@jit
def Bfunc(kFsat,satDens,gSigma,gOmega,gRho,b,c,nr):
    epsilon = nuclearMatterEoS(kFsat,gSigma,gOmega,gRho,b,c,nr,False)[1]
    return epsilon/satDens-m

###################################################
#Returns the criteria that dB/drho=0 at saturation#
###################################################
@jit
def densFunc(kFsat,satDens,dkF,dRho,gSigma,gOmega,gRho,b,c,nr):
    epsilon = nuclearMatterEoS(kFsat,gSigma,gOmega,gRho,b,c,nr,False)[1]
    epsilonPlus = nuclearMatterEoS(kFsat+dkF,gSigma,gOmega,gRho,b,c,nr,False)[1]
    epsilonMinus = nuclearMatterEoS(kFsat-dkF,gSigma,gOmega,gRho,b,c,nr,False)[1] 
    return (satDens)**(-1)*(epsilonPlus-epsilonMinus)**(-1.)*dRho*epsilon

#################################
#Returns the compression modolus#
#################################
@jit
def Kfunc(kFsat,satDens,dkF,dRho,gSigma,gOmega,gRho,b,c,nr):
    epsilon = nuclearMatterEoS(kFsat,gSigma,gOmega,gRho,b,c,nr,False)[1]
    epsilonPlus = nuclearMatterEoS(kFsat+dkF,gSigma,gOmega,gRho,b,c,nr,False)[1]
    epsilonMinus = nuclearMatterEoS(kFsat-dkF,gSigma,gOmega,gRho,b,c,nr,False)[1] 
    epsilonDensPlus = epsilonPlus/(2*density(kFsat+dkF))
    epsilonDens = epsilon/satDens
    epsilonDensMinus = epsilonMinus/(2*density(kFsat-dkF))
    return kFsat**2*(epsilonDensPlus+epsilonDensMinus-2*epsilonDens)*dkF**(-2)

#########################################################################
#Function that returns the values of computed binding energy and        # 
#saturation density for a given set of couplings and the empirical value#
#########################################################################
def bindingEnergyAndSaturationDensityCheck(kFvec,satDens,satB,nuclearEpsilonVec,N):
    k = kFvec[0]
    dens = 2*density(k[1:N])
    nuclearEpsilonVec = nuclearEpsilonVec[1:N]
    N=N-1
    epsilon = interpolate.interp1d(dens,nuclearEpsilonVec)
    func = lambda den: epsilon(den)/den - m   
    satDensIndex = list(func(dens)).index(min(func(dens)))
    B = min(func(dens))
    satDens = MeVtoFm(satDens*939**3,3)
    return satDensIndex,[MeVtoFm(dens[satDensIndex]*939**3,3),satDens],[B*939,satB*939]

#############################################################
#Function that returns the values of computed effective mass# 
#for a given set of couplings and the empirical value       #
#############################################################
def mStarCheck(kFvec,satMstar,nuclearMstarVec,satDensIndex,N):
    return [np.float(nuclearMstarVec[satDensIndex]), satMstar]
    
##################################################################
#Function that returns the values of computed compression modolus# 
#for a given set of couplings and the empirical value            #
##################################################################
def compressionModuloCheck(kFvec,satK,nuclearEpsilonVec,satDensIndex):
    k = kFvec[0]
    deltaEpsilon = nuclearEpsilonVec[satDensIndex+1]-nuclearEpsilonVec[satDensIndex-1]
    deltaEpsilon2 = nuclearEpsilonVec[satDensIndex+1]+nuclearEpsilonVec[satDensIndex-1]
    deltaK = k[satDensIndex+1]-k[satDensIndex-1]
    derEpsilon = deltaEpsilon/deltaK
    derEpsilon2 = (deltaEpsilon2-2*nuclearEpsilonVec[satDensIndex])*(0.5*deltaK)**(-2)
    satDens = 2*density(k[satDensIndex])
    satkF = k[satDensIndex]
    satE = nuclearEpsilonVec[satDensIndex]
    mem1 = derEpsilon2
    if(satkF == 0.):
        K = 0.
    else:
        mem2 = 6*derEpsilon/satkF
        mem3 = 6*satE*satkF**(-2)
        mem4 = 18*satE*satkF**(-2)
        K = satkF**2*(mem1-mem2-mem3+mem4)*satDens**(-1)
    return [K*939,satK*939]

##########################################################################
#Function that returns the values of computed symmetry energy coefficient# 
#for a given set of couplings and the empirical value                    #
##########################################################################
def symmetryCoefficientCheck(kFvec,satAs,satDensIndex,mStar,gRho):
    k = kFvec[0]
    satkF = k[satDensIndex]
    if(satkF==0.):
        S = 0.
    else:
        mem1 = gRho**2*mRho**(-2)*satkF**3/(12*math.pi**2)
        mem2 = satkF**2/(6*(satkF**2+mStar**2)**(0.5))
        S = (mem1+mem2)*939
    return [S,satAs*939]
    
#############################################################
#Returns the results from all the four check-functions above#
#############################################################
def runAllChecks(kFvec,satDens,satB,satK,satMstar,satAs,nuclearMstarVec,nuclearEpsilonVec,gRho):
    N = len(nuclearEpsilonVec)
    satDensIndex,B,D = bindingEnergyAndSaturationDensityCheck(kFvec,satDens,satB,nuclearEpsilonVec,N)
    mStar = mStarCheck(kFvec,satMstar,nuclearMstarVec,satDensIndex,N)
    K = compressionModuloCheck(kFvec,satK,nuclearEpsilonVec,satDensIndex)
    symmetryCoefficientCheck(kFvec,satAs,satDensIndex,mStar[0],gRho)
    return B,D,mStar,K

###################################
#Calculates the coupling constants#
###################################
def couplingConstants(best,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,N,tolerance,nr,hyp):    
    kFsat = momentaFromDensity(satDens)/2**(1./3)
    dkF = 10**(-5)
    if(N!=1):
        deltaSigma = (gSigmaRange[1]-gSigmaRange[0])/(N-1)
        deltaOmega = (gOmegaRange[1]-gOmegaRange[0])/(N-1)
        deltab = (bRange[1]-bRange[0])/(N-1)
        deltac = (cRange[1]-cRange[0])/(N-1)
    else: 
        deltaSigma = (gSigmaRange[1]-gSigmaRange[0])/2
        deltaOmega = (gOmegaRange[1]-gOmegaRange[0])/2
        deltab = (bRange[1]-bRange[0])/2
        deltac = (cRange[1]-cRange[0])/2
    gSigmaGuess = gSigmaRange[0]-deltaSigma
    gOmegaGuess = gOmegaRange[0]-deltaOmega
    bGuess = bRange[0]-deltab
    cGuess = cRange[0]-deltac
    dRho = 2*(density(kFsat+dkF)-density(kFsat-dkF))
    gRho = findgRho(kFsat,satMstar,satAs)
    func = lambda arg:[(satMstar-mStarFuncCouplings(kFsat,arg[0],arg[1],gRho,arg[2],arg[3],nr)),
                        (satB-Bfunc(kFsat,satDens,arg[0],arg[1],gRho,arg[2],arg[3],nr)),
                        (densFunc(kFsat,satDens,dkF,dRho,arg[0],arg[1],gRho,arg[2],arg[3],nr)),
                        (satK-Kfunc(kFsat,satDens,dkF,dRho,arg[0],arg[1],gRho,arg[2],arg[3],nr))]
    bestgSigma = gSigmaGuess+0.5*(gSigmaRange[1]-gSigmaRange[0])
    bestgOmega = gOmegaGuess+0.5*(gOmegaRange[1]-gOmegaRange[0])
    bestb = bGuess+0.5*(bRange[1]-bRange[0])
    bestc = cGuess+0.5*(cRange[1]-cRange[0])
    solver = 'hybr'
    for i in range(N):
        gSigmaGuess+=deltaSigma
        for j in range(N):
            gOmegaGuess+=deltaOmega
            for l in range(N):
                bGuess+=deltab
                for n in range(N):
                    print((N**4),(n+l*N+j*N*N+i*N*N*N+1),best)
                    cGuess+=deltac
                    solution = op.root(func,[gSigmaGuess,gOmegaGuess,bGuess,cGuess],method=solver)
                    gSigma,gOmega,b,c = solution.x
                    if (solution.success == True and (b<0) and (c<0)):
                        N2 = 100
                        kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear = createVectors(1.,0.03,[gSigma,gSigma],[gOmega,gOmega],[gRho,gRho],[0,0,0],[0,0,0],[0,0,0],[b,b],[c,c],N2,nr,hyp)
                        d,e,f,g = runAllChecks(kFvec,satDens,satB,satK,satMstar,satAs,nuclear[0],nuclear[1],gRho)
                        d = 1.-d[0]/d[1]
                        e = 1.-e[0]/e[1]
                        f = 1.-f[0]/f[1]
                        g = 1.-g[0]/g[1]
                        test = np.sqrt(d**2+e**2+f**2+g**2)
                        if(test<best):
                            N2 = 5000
                            kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear = createVectors(1.,0.03,[gSigma,gSigma],[gOmega,gOmega],[gRho,gRho],[0,0,0],[0,0,0],[0,0,0],[b,b],[c,c],N2,nr,hyp)
                            d,e,f,g = runAllChecks(kFvec,satDens,satB,satK,satMstar,satAs,nuclear[0],nuclear[1],gRho)
                            d = 1.-d[0]/d[1]
                            e = 1.-e[0]/e[1]
                            f = 1.-f[0]/f[1]
                            g = 1.-g[0]/g[1]
                            test = np.sqrt(d**2+e**2+f**2+g**2)
                            if(test<best):
                                best = test
                                print(best)
                                if(best<tolerance):
                                    print("Coupling sucess for",nr)
                                    return gSigma,gOmega,gRho,b,c,best
                            bestgSigma = gSigma
                            bestgOmega = gOmega
                            bestb = b
                            bestc = c
                cGuess = cRange[0]-deltac
            bGuess = bRange[0]-deltab
        gOmegaGuess = gOmegaRange[0]-deltaOmega
    print("Failed to find satisfactory couplings for",nr)
    print(bestgSigma,bestgOmega,gRho,bestb,bestc,best)
    return bestgSigma,bestgOmega,gRho,bestb,bestc,best
                        
                        
#####################################################
#Creates coupling constants, then write them to file#
#####################################################
def writeCouplingConstantsToFile(satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,N,iterations,multiplier,tolerance,nr,hyp,filename):
    best = 1000
    for i in range(iterations-1): 
        gSigmatest,gOmegatest,gRhotest,btest,ctest,test = couplingConstants(best,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,N,tolerance,nr,hyp)
        if(test<best):
            best=test
            gSigma = gSigmatest
            gOmega = gOmegatest
            gRho = gRhotest
            b = btest
            c = ctest
            gSigmaRange[1] = gSigma+(gSigmaRange[1]-gSigmaRange[0])/2*multiplier
            gSigmaRange[0] = gSigma-(gSigmaRange[1]-gSigmaRange[0])/2*multiplier
            gOmegaRange[1] = gOmega+(gOmegaRange[1]-gOmegaRange[0])/2*multiplier
            gOmegaRange[0] = gOmega-(gOmegaRange[1]-gOmegaRange[0])/2*multiplier
            bRange[1] = b+(bRange[1]-bRange[0])/2*multiplier
            bRange[0] = b-(bRange[1]-bRange[0])/2*multiplier
            cRange[1] = c+(cRange[1]-cRange[0])/2*multiplier
            cRange[0] = c-(cRange[1]-cRange[0])/2*multiplier
        else:
            break
        if(best<tolerance):
            gSigma = gSigmatest
            gOmega = gOmegatest
            gRho = gRhotest
            b = btest
            c = ctest
            if(nr=='r'):
                filename = filename+'Renormalized'
            f = open(filename,'w')
            f.write(str(gSigma)+" "+str(gOmega)+" "+str(gRho)+" "+str(b)+" "+str(c)+"\n")
            f.close()
            return
        print('-----------------',i,'-------------------')
        
    print('Falure couplings')
    best=test
    gSigma = gSigmatest
    gOmega = gOmegatest
    gRho = gRhotest
    b = btest
    c = ctest
    if(nr=='r'):
        filename = filename+'Renormalized'
    f = open(filename,'w')
    f.write(str(gSigma)+" "+str(gOmega)+" "+str(gRho)+" "+str(b)+" "+str(c)+"\n")
    f.close()
    return
    return  
    
    
##################################################################
#Takes a filename and returns the coupling constants in that file#
##################################################################    
def readCouplingConstantsFromFile(filename,mSigma):
    numbers = np.zeros(5,np.float)
    if(mSigma==550./939):
        f = open(filename,'r')
        line = f.readline()
        numbers[:] = line.split() 
        f.close()
    f = open(filename+'Renormalized','r')
    line = f.readline()
    numbersR = np.zeros(5,np.float)
    numbersR[:] = line.split() 
    f.close()
    return ([numbers[0],numbersR[0]],[numbers[1],numbersR[1]],[numbers[2],numbersR[2]],
    [numbers[3],numbersR[3]],[numbers[4],numbersR[4]])


############################################################
#Creates all vectors (Fermi momenta, effective masses etc) #
############################################################
@jit    
def createVectors(keMax,densMax,gSigma,gOmega,gRho,xSigma,xOmega,xRho,b,c,N,nr,hyp):
    if(nr=='n'):
        gSigma = gSigma[0]
        gOmega = gOmega[0]
        gRho = gRho[0]
        b = b[0]
        c = c[0]
    else:
        gSigma = gSigma[1]
        gOmega = gOmega[1]
        gRho = gRho[1]
        b = b[1]
        c = c[1]
    if(hyp==True):
        densVec = np.linspace(0,densMax,N+1)
        keVec = np.zeros(N+1)
    else:
        keVec = np.linspace(0,keMax,N+1)
        densVec = np.zeros(N+1)
        
    kMuVec = np.zeros(N+1)
    knVec = np.zeros(N+1)
    kpVec = np.zeros(N+1)
    klambdaVec = np.zeros(N+1)
    ksminusVec = np.zeros(N+1)
    ks0Vec = np.zeros(N+1)
    ksplusVec = np.zeros(N+1)
    kximinusVec = np.zeros(N+1)
    kxi0Vec = np.zeros(N+1)
    omegaVec = np.zeros(N+1)
    rhoVec = np.zeros(N+1)
    mStarVec = np.ones(N+1)
    mStarLambdaVec = np.ones(N+1)*mLambda
    mStarSvec = np.ones(N+1)*mS
    mStarXiVec = np.ones(N+1)*mXi
    Pvec = np.zeros(N+1)
    epsilonVec = np.zeros(N+1)
    nuclearEpsilon = np.zeros(N+1)
    nuclearP = np.zeros(N+1)
    nuclearMstar = np.ones(N+1)
    counter = 0
    
    knVec[0] = (densVec[0]*3*math.pi**2)**(1./3)
    for i in range(1,N+1):
        if(hyp==True):
            guesses = [omegaVec[i-1],rhoVec[i-1],keVec[i-1],knVec[i-1],mStarVec[i-1]]
            omegaVec[i],rhoVec[i],keVec[i],kMuVec[i],knVec[i],kpVec[i],klambdaVec[i],ksminusVec[i],ks0Vec[i],ksplusVec[i],kximinusVec[i],kxi0Vec[i],mStarVec[i],mStarLambdaVec[i],mStarSvec[i],mStarXiVec[i],sol = fermiMomentaAndEffectiveMass(densVec[i],guesses,gSigma,gOmega,gRho,xSigma,xOmega,xRho,b,c,nr,hyp)
            kFvec = [keVec[i],kMuVec[i],knVec[i],kpVec[i],klambdaVec[i],ksminusVec[i],ks0Vec[i],ksplusVec[i],kximinusVec[i],kxi0Vec[i]]
        else:
            guesses = [omegaVec[i-1],rhoVec[i-1],knVec[i-1],mStarVec[i-1]]
            omegaVec[i],rhoVec[i],keVec[i],kMuVec[i],knVec[i],kpVec[i],klambdaVec[i],ksminusVec[i],ks0Vec[i],ksplusVec[i],kximinusVec[i],kxi0Vec[i],mStarVec[i],mStarLambdaVec[i],mStarSvec[i],mStarXiVec[i],sol = fermiMomentaAndEffectiveMass(keVec[i],guesses,gSigma,gOmega,gRho,xSigma,xOmega,xRho,b,c,nr,hyp)
            kFvec = [keVec[i],kMuVec[i],knVec[i],kpVec[i],klambdaVec[i],ksminusVec[i],ks0Vec[i],ksplusVec[i],kximinusVec[i],kxi0Vec[i]]
            dens = density(keVec[i]) + density(kMuVec[i]) + density(knVec[i])
            dens = dens + density(kpVec[i])+density(klambdaVec[i]) + density(ksminusVec[i])
            dens = dens + density(ks0Vec[i])+density(ksplusVec[i]) + density(kximinusVec[i]) + density(kxi0Vec[i])
            densVec[i] = dens
        if(sol==True):
                counter += 1        
        
        nucM,nuce,nucP = nuclearMatterEoS(keVec[i],gSigma,gOmega,gRho,b,c,nr,hyp)
        nuclearMstar[i] = nucM
        nuclearEpsilon[i] = nuce
        nuclearP[i] = nucP
        Pvec[i] = pressure(kFvec,mStarVec[i],mStarLambdaVec[i],mStarSvec[i],mStarXiVec[i],gSigma,gOmega,gRho,b,c,xRho,nr,hyp)
        epsilonVec[i] = energyDensity(kFvec,mStarVec[i],mStarLambdaVec[i],mStarSvec[i],mStarXiVec[i],gSigma,gOmega,gRho,b,c,xRho,nr,hyp)
    print('FermiMomentaAndEffectiveMass converged',counter,'out of',N,'times')
    kFvec = [keVec,kMuVec,knVec,kpVec,klambdaVec,ksminusVec,ks0Vec,ksplusVec,kximinusVec,kxi0Vec]
    mStarVec = [mStarVec,mStarLambdaVec,mStarSvec,mStarXiVec]
    nuclearVec = [nuclearMstar,nuclearEpsilon,nuclearP]
    return kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclearVec





################################
###                          ###
###   Mass-radii relations   ###   
###                          ###
################################
       
###########################################################
#The derivative of the mass M with respect to the radius r#
#for arbitrary relativity                                 #
###########################################################
@jit
def dMdr(r,epsilon):
    return beta*epsilon*r**2

###############################################################
#The derivative of the pressure P with respect to the radius r#
#for arbitrary relativity                                     #
###############################################################
@jit
def dPdrTOV(r,P,M,epsilon):
    return-R0*r**(-2.)*(epsilon+P)*(M+beta*P*r**3)*(1.-2.*R0*M*r**(-1))**(-1)


###################################################################
#Function that creates the coefficients in the Runge Kutta routine# 
###################################################################
@jit
def k(r,P,M,epsilon,h):
    kP1 = dPdrTOV(r,P,M,epsilon)
    kM1 = dMdr(r,epsilon)
    kP2 = dPdrTOV(r+h/2,P+h/2*kP1,M+h/2*kM1,epsilon)
    kM2 = dMdr(r+h/2,epsilon)
    kP3 = dPdrTOV(r+h/2,P+h/2*kP2,M+h/2*kM2,epsilon)
    kM3 = dMdr(r+h/2,epsilon)
    kP4 = dPdrTOV(r+h/2,P+h*kP3,M+h*kM3,epsilon)
    kM4 = dMdr(r+h,epsilon)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4


#############################################################
#Function that returns 2 vectors containing the mass-radius #
#relation and a flag indicating if the maximum number of    #
#iterations is reached                                      # 
#############################################################
@jit
def results(Pc,h,nMax,EoS,PvecMin):
    P = Pc
    M = 0.
    r = 0.
    flag=True
    for i in range(nMax):
        if(flag==True):
            Memoryr = r
            r = r+h
            MemoryM = M
            if(P<PvecMin):
                epsilon=0.
                print(P)
            else:
                epsilon = EoS(P)
            kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4 = k(r,P,M,epsilon,h)
            P = P+h/6*(kP1+2*kP2+2*kP3+kP4)
            M = M+h/6*(kM1+2*kM2+2*kM3+kM4)
            if(np.float(np.real(P))<=0. or M==MemoryM):
                r = Memoryr
                flag = False
                break

    if(flag == True):
            print("Maximum number of iterations reached")
            print("for TOV-equation with Pc = %.8f"%Pc)

    return r, np.float(np.real(M)),flag
  
        
###############################################################
#Does the same as parametrisingNonRel for arbitrary relativity#
###############################################################
@jit    
def paramterising(parameters,epsilonVec,Pvec):
    PcMmax = 0.
    PcMin = parameters[0]
    PcMax = parameters[1]
    N1 = int(parameters[2])
    h = parameters[3]
    nMax = int(parameters[4])
    N2 = len(Pvec)
    if(PcMax>Pvec[N2-1]):
        print("PcMax changed from ",PcMax," to ", Pvec[N2-1])
        PcMax = Pvec[N2-1]
    PvecMin = min(Pvec)  
    Pc = PcMin
    const = (np.float(PcMax)/PcMin)**(1./N1)
    R = np.zeros(N1,np.double)
    M = np.zeros(N1,np.double)

    EoS = interpolate.interp1d(Pvec,epsilonVec)
    mMax = 0.
    for i in range(N1):
        print(i)
        R[i],M[i],flag = results(Pc,h,nMax,EoS,PvecMin)
        if(M[i]>mMax):
            PcMmax = Pc
        if(flag==True):
            print(R[i])
            R = R[0:i]
            M = M[0:i]
            N1 = i
            break
        Pc = Pc*const
    print("Pressure in the center of maximum mass:",PcMmax) #Print the pressure in the star with largest mass
    return R,M,N1
       
#####################################
#Writes mass-radius relation to file#
#####################################
def writeResultsToFile(parameters,epsilonVec,Pvec,filename,nr,hyp):
    if(nr=='r'):
        filename = filename+'Renormalized'
    elif(hyp==True):
        filename = filename+'Hyperon'
    R2,M2,N = paramterising(parameters,epsilonVec,Pvec)
    
    f = open(filename,'w')
    f.write(str(N)+'\n')
    for i in range(N):
        a = str(R2[i])
        b = str(M2[i])
        f.write(a+" "+b+"\n")
    f.close()
    return  

########################################
#Writes all data used in theses to file#
########################################
def writeAllThingsToFile(keMax,densMax,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,filenameCouplings,filenameEoS,filenameMassRadii,mSigma,N,N2):
    if(mSigma==550./939):
        N3=2
    else:
        N3=1
    for i in range(N3):
        if(i==1):
            nr = 'n'
        else:
            nr = 'r'
        nr='n'
        writeCouplingConstantsToFile(satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,N,10,1./N,0.05,nr,False,filenameCouplings)
    gSigma,gOmega,gRho,b,c= readCouplingConstantsFromFile(filenameCouplings,mSigma)
    if(N3==2):
        N3=3
    for i in range(N3):
        if(i==2):
            nr = 'n'
            hyp = False
        elif(i==0):
            nr = 'r'
            hyp = False
        else:
            nr = 'n'
            hyp = True 
        writeEoStoFile(keMax,densMax,gSigma,gOmega,gRho,b,c,N2,filenameEoS,nr,hyp)
        kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenameEoS,nr,hyp)
        writeResultsToFile(parameters,epsilonVec,Pvec,filenameMassRadii,nr,hyp)
        

#######################################
#Reads mass-radius relations from file#
#######################################
def readResultsFromFile(filename,nr,hyp):
    if(nr=='r'):
        filename = filename+'Renormalized'
    elif(hyp==True):
        filename = filename+'Hyperon'
    print(filename)
    f = open(filename,'r')
    N = int(f.readline())
    R = np.zeros(N,np.double)
    M = np.zeros(N,np.double)
    i = 0
    data = f.readlines()
    for line in data:
        numbers = line.split()
        R[i] = numbers[0]
        M[i] = numbers[1]
        i = i+1 
    f.close()
    return R,M,N

########################
#Writes the EoS to file#
########################
def writeEoStoFile(keMax,densMax,gSigma,gOmega,gRho,b,c,N,filename,nr,hyp):
    if(nr=='r'):
        filename=filename+'Renormalized'
    elif(hyp==True):
        filename=filename+'Hyperon'
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear = createVectors(keMax,densMax,gSigma,gOmega,gRho,xSigma,xOmega,xRho,b,c,N,nr,hyp)
    ke = kFvec[0]
    kMu = kFvec[1]
    kn = kFvec[2]
    kp = kFvec[3]
    klambda = kFvec[4]
    ksminus = kFvec[5]
    ks0 = kFvec[6]
    ksplus = kFvec[7]
    kximinus = kFvec[8]
    kxi0 = kFvec[9]
    nucMstar = nuclear[0]
    nucEpsilon = nuclear[1]
    nucP = nuclear[2]
    mStar = mStarVec[0]
    mStarLambda = mStarVec[1]
    mStarS = mStarVec[2]
    mStarXi = mStarVec[3]
    f = open(filename,'w')
    f.write(str(N)+'\n')
    for i in range(N):
        a = str(ke[i])
        b = str(kMu[i])
        c = str(kn[i])
        d = str(kp[i])
        e = str(klambda[i])
        g = str(ksminus[i])
        h = str(ks0[i])
        j = str(ksplus[i])
        k = str(kximinus[i])
        l = str(kxi0[i])
        m = str(densVec[i])
        n = str(mStar[i])
        o = str(mStarLambda[i])
        p = str(mStarS[i])
        q = str(mStarXi[i])
        r = str(epsilonVec[i])
        s = str(Pvec[i])
        t = str(nucMstar[i])
        u = str(nucEpsilon[i])
        v = str(nucP[i]) 
        
        f.write(a+" "+b+" "+c+" "+d+" "+e+" "+g+" "+h+" "+j+" "+k+" "+l+" "+m+" ")
        f.write(n+" "+o+" "+p+" "+q+" "+r+" "+s+" "+t+" "+u+" "+v+"\n")
    f.close()
    return   

#########################
#Reads the EoS from file#
#########################
def readEoSfromFile(filename,nr,hyp):
    if(nr=='r'):
        filename=filename+'Renormalized'
    elif(hyp==True):
        filename = filename+'Hyperon'
    f = open(filename,'r')
    N = int(f.readline())
    ke = np.zeros(N)
    kMu = np.zeros(N)
    kn = np.zeros(N)
    kp = np.zeros(N)
    klambda = np.zeros(N)
    ksminus = np.zeros(N)
    ks0 = np.zeros(N)
    ksplus = np.zeros(N)
    kximinus = np.zeros(N)
    kxi0 = np.zeros(N)
    densVec = np.zeros(N)
    mStarVec = np.zeros(N)
    mStarLambdaVec = np.zeros(N)
    mStarSvec = np.zeros(N)
    mStarXiVec = np.zeros(N)
    epsilonVec = np.zeros(N)
    Pvec = np.zeros(N)
    nucMstar = np.zeros(N)
    nucEpsilon = np.zeros(N)
    nucP = np.zeros(N)
    i = 0
    data = f.readlines()
    for line in data:
        numbers = line.split()
        ke[i] = numbers[0]
        kMu[i] = numbers[1]
        kn[i] = numbers[2]
        kp[i] = numbers[3]
        klambda[i] = numbers[4]
        ksminus[i] = numbers[5]
        ks0[i] = numbers[6]
        ksplus[i] = numbers[7]
        kximinus[i] = numbers[8]
        kxi0[i] = numbers[9]
        densVec[i] = numbers[10]
        mStarVec[i] = numbers[11]
        mStarLambdaVec[i] = numbers[12]
        mStarSvec[i] = numbers[13]
        mStarXiVec[i] = numbers[14]
        epsilonVec[i] = numbers[15]
        Pvec[i] = numbers[16]
        nucMstar[i] = numbers[17]
        nucEpsilon[i] = numbers[18]
        nucP[i] = numbers[19]
        i = i+1 
    return [ke,kMu,kn,kp,klambda,ksminus,ks0,ksplus,kximinus,kxi0],densVec,[mStarVec,mStarLambdaVec,mStarSvec,mStarXiVec],epsilonVec,Pvec,[nucMstar,nucEpsilon,nucP],N
        
#################################
#Plots all mass radius relations#
#################################
def plotMassRadiusRelation(filenamesMassRadii):
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.yaxis.set_label_coords(-0.05, 1)
    ax.set_xlabel('$R$[km]',fontsize = 13)
    ax.set_ylabel('$M/M_\odot$', rotation='horizontal', fontsize = 13)
    colours = ['b','y','r','g','m']
    for i in range(3):
        mSigma = 500 + 50*i
        if(i==0):
            R,M,N = readResultsFromFile(filenamesMassRadii[1],'n',False)
            maxIndex = list(M).index(max(list(M)))
            print("npmue-matter: Maximum mass is",M[maxIndex],"solar masses with radius", R[maxIndex], "km\n")
            plt.plot(R[0:maxIndex+1],M[0:maxIndex+1],colours[0],label = 'MFA')
            plt.plot(R[maxIndex:N],M[maxIndex:N],colours[0]+'--')
            R,M,N = readResultsFromFile(filenamesMassRadii[1],'n',True)
            maxIndex = list(M).index(max(list(M)))
            print("hyperon-matter: Maximum mass is",M[maxIndex],"solar masses with radius", R[maxIndex], "km\n")
            plt.plot(R[0:maxIndex+1],M[0:maxIndex+1],colours[4],label = 'MFA with Hyperons')
            plt.plot(R[maxIndex:N],M[maxIndex:N],colours[4]+'--')
            
        R,M,N = readResultsFromFile(filenamesMassRadii[i],'r',False)
        maxIndex = list(M).index(max(list(M)))
        print("Renormalized npmue-matter, mSigma =",mSigma,": Maximum mass is",M[maxIndex],"solar masses with radius", R[maxIndex], "km\n")
        
        plt.plot(R[0:maxIndex+1],M[0:maxIndex+1],colours[i+1],label = 'RHA with $m_\sigma=$'+str(mSigma))
        plt.plot(R[maxIndex:N],M[maxIndex:N],colours[i+1]+'--')
    plt.legend()
    
#######################################
#Plots the equation of state from file#
#######################################
def plotEoS(filenameEoS):
    plt.figure()
    plt.xlim(0,0.02)
    plt.ylim(0,0.01)
    ax = plt.gca()
    ax.xaxis.set_label_coords(1., -0.07)
    ax.set_xlabel('$\\bar{\epsilon}$', fontsize = 15)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 15)
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'n',False)
    plt.plot(epsilonVec,Pvec,'b',label='MFA')
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'r',False)
    plt.plot(epsilonVec,Pvec,'r',label='RHA')
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'n',True)
    plt.plot(epsilonVec,Pvec,'m',label='MFA with hyperons')
    plt.legend()

 
#####################################################################
#Plots the population density from using the file containing the EoS#
#####################################################################
def plotPopulationDensity(filenameEoS):
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'n',False)
    plt.figure()
    ax = plt.gca()
    plt.xlim(0,1.6)
    plt.ylim(0.001,1.01)
    ax.text(0.84,0.66,'n',fontsize=14)
    ax.text(0.60,0.19,'p',fontsize=14)
    ax.text(0.57,0.105,'e',fontsize=14)
    ax.text(0.12,0.0066,'$\mu$',fontsize=14)
    
    ax.xaxis.set_label_coords(1., -0.05)
    ax.set_xlabel('$\\rho$[fm$^{-3}$]', fontsize = 15)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\rho_i/\\rho$', rotation='horizontal',fontsize = 15)
    
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],density(kFvec[0])[1:N2+1]/densVec[1:N2+1],'k',label='e')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[1])[1:N2+1]/densVec[1:N2+1]),'r',label='$\mu$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[2])[1:N2+1]/densVec[1:N2+1]),'b',label='n')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[3])[1:N2+1]/densVec[1:N2+1]),'g',label='p')   
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'n',True)
    
    
    plt.figure()
    ax = plt.gca()
    plt.xlim(0,1.6)
    plt.ylim(0.001,1.01)
    ax.text(0.616,0.55,'n',fontsize=14)
    ax.text(0.596,0.25,'p',fontsize=14)
    ax.text(0.34,0.10,'e',fontsize=14)
    ax.text(0.10,0.0066,'$\mu$',fontsize=14)
    ax.text(0.24,0.0013,'$\Sigma^-$',fontsize=14)
    ax.text(0.36,0.007,'$\Lambda$',fontsize=14)
    ax.text(0.586,0.0013,'$\Xi^-$',fontsize=14)
    ax.text(0.83,0.007,'$\Sigma^0$',fontsize=14)
    ax.text(0.92,0.0013,'$\Sigma^+$',fontsize=14)
    ax.text(1.14,0.007,'$\Xi^0$',fontsize=14)
    
    ax.xaxis.set_label_coords(1., -0.05)
    ax.set_xlabel('$\\rho$[fm$^{-3}$]', fontsize = 15)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\rho_i/\\rho$', rotation='horizontal',fontsize = 15)
    
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],density(kFvec[0])[1:N2+1]/densVec[1:N2+1],'k',label='e')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[1])[1:N2+1]/densVec[1:N2+1]),'r',label='$\mu$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[2])[1:N2+1]/densVec[1:N2+1]),'b',label='n')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[3])[1:N2+1]/densVec[1:N2+1]),'g',label='p')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[4])[1:N2+1]/densVec[1:N2+1]),'k--',label='$\Lambda$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[5])[1:N2+1]/densVec[1:N2+1]),'r--',label='$\Sigma^-$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[6])[1:N2+1]/densVec[1:N2+1]),'b--',label='$\Sigma^0$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[7])[1:N2+1]/densVec[1:N2+1]),'g--',label='$\Sigma^+$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[8])[1:N2+1]/densVec[1:N2+1]),'y--',label='$\Xi^-$')
    plt.semilogy(MeVtoFm(densVec*939**3,3)[1:N2+1],(density(kFvec[9])[1:N2+1]/densVec[1:N2+1]),'m--',label='$\Xi^0$')       

##########################
#Plots the binding energy#
##########################
def plotBindingEnergy(kFvec,epsilonVec,nuclearEpsilonVec):
    dens = density(kFvec[2])+density(kFvec[3])
    nuclearDens = 2*density(kFvec[0])
    N = len(epsilonVec)
    B = epsilonVec[1:N]/dens[1:N]-m
    nuclearB = nuclearEpsilonVec[1:N]/nuclearDens[1:N]-m
    plt.figure()
    plt.xlim(0,1.5)
    plt.ylim(-50,300)
    plt.plot(MeVtoFm(dens[1:N]*939**3,3),B*939,'b')
    plt.plot(MeVtoFm(nuclearDens[1:N]*939**3,3),nuclearB*939,'r')
    return

###################################################################
#Takes in the filenames of all relevant files and plots everything#
###################################################################
def plotEverything(filenamesEoS,filenamesMassRadii):
    kFvec,densVec,mStarVec,epsilonVec,Pvec,nuclear,N2 = readEoSfromFile(filenamesEoS[1],'nr',False)
    plotBindingEnergy(kFvec,epsilonVec,nuclear[1])
    plotEoS(filenamesEoS[1])
    plotPopulationDensity(filenamesEoS[1])  
    plotMassRadiusRelation(filenamesMassRadii)


#########################        
###                   ###
###   Program start   ###
###                   ###
#########################   

#Some spesifications:

#The variable 'hyp' is bool. If hyp==True then calculations include hyperons
#The variable 'nr' is string. If nr=='n' we use MFA, while nr='r' uses RHA

############
#Parameters#
############

#particle masses
m = 1.
me = 0.51099/939
mMu = 105.7/939.
mSigma = 550./939
mOmega = 783./939
mRho = 775.5/939
mLambda = 1115./939
mS = 1192./939
mXi = 1318./939

#Other parameters
R0 = 1.47
beta = 1.1426

#Nuclear matter properties
landauMass = 0.83
satDens = fmToMeV(0.153,-3)/939**3
satMstar =  np.sqrt(landauMass**2-((3/2*math.pi**2*satDens)**(1./3))**2)
satB = -16.3/939
satK = 300./939
satAs = 32.5/939

#Variables used when computing massradius relations
PcMin = 4*10**(-6)     #Minimum central pressure
PcMax = 1.2*10**(0)    #Maximum central pressure
N1 = 300               #Number of data points in mass-radius plot
h = 0.001              #Step size in km
nMax = 100/h           #Maximum number of iterations

#Collects the mass-radius parameters in one vector
parameters = [PcMin,PcMax,N1,h,nMax]

N = 8#Number of intervalls used to compute the coupling constants

N2 = 50000         #Number of data points used for the EoS
keMax = 0.5        #Maximum electron Fermi momentum
densMax = 0.035    #Maximum baryon density

#Coupling ratios: [xgLambda,xgSigma,xgXi]
xSigma = [1,1,1]      
xOmega = [1,1,1]
xRho = [1,1,1]

#Filenames
filenamesCouplings = ['couplingconstants500','couplingconstants550','couplingconstants600']
filenamesEoS = ['equationOfState500','equationOfState550','equationOfState600']
filenamesMassRadii = ['massRadiusRelation500','massRadiusRelation550','massRadiusRelation600']

#The ranges we look for the coupling constants
gSigmaRange =  [6,7]
gOmegaRange =  [8,9]
bRange = [-0.01,0]
cRange = [-0.01,0]
mSigma = 500./939
writeAllThingsToFile(keMax,densMax,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,filenamesCouplings[0],filenamesEoS[0],filenamesMassRadii[0],mSigma,N,N2)
print(1,'out of',3,'complete!')
gSigmaRange =  [7,9]
gOmegaRange =  [8,9]
bRange = [-0.01,0]
cRange = [-0.01,0]
mSigma = 550./939
writeAllThingsToFile(keMax,densMax,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,filenamesCouplings[1],filenamesEoS[1],filenamesMassRadii[1],mSigma,N,N2)
print(2,'out of',3,'complete! Nearly there...')
gSigmaRange =  [9,10]
gOmegaRange =  [8,9]
bRange = [-0.01,0]
cRange = [-0.01,0]
mSigma = 600./939
writeAllThingsToFile(keMax,densMax,satDens,satMstar,satB,satK,satAs,gSigmaRange,gOmegaRange,bRange,cRange,filenamesCouplings[2],filenamesEoS[2],filenamesMassRadii[2],mSigma,N,N2)
print('Done! Just plotting now...')

plotEverything(filenamesEoS,filenamesMassRadii)


plt.show()



print("\nTime spent:")
print(time.clock()-t0)