import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as integrate
import time
from numba import jit
from scipy import optimize as op
import math
from scipy import interpolate 

##########################################################
##########################################################
###                                                    ###
###   Just press play to create all plots in thesis!   ###
###                                                    ###
##########################################################
##########################################################

#Variable used to time the program
t0 = time.clock()         

###########################
###                     ###
###  Defining functons  ###
###                     ###
###########################

#################################################
#Takes in a number in fm and converts it to MeV #
#Takes in the exponent. For instance, to convert#
#from fm^-3 to MeV^3, set exponent to -3        #
#################################################
def fmToMeV(number,exponent):
    if(exponent>0):
        return (number**(1./exponent)/197.33)**exponent
    else:
        return (number**(-1./exponent)*197.33)**(-exponent)

#################################################
#Takes in a number in MeV and converts it to fm #
#Takes in the exponent. For instance, to convert#
#from MeV^3 to fm^-3, set exponent to 3         #
#################################################
def MeVtoFm(number,exponent):
    if(exponent>0):
        return (number**(1./exponent)/197.33)**exponent
    else:
        return (number**(-1./exponent)*197.33)**(-exponent)

################################################################
#Function that takes in a density and returns the corresponding#
#Fermi momentum given a degenracy factor f                     #
################################################################
@jit
def kFfromDensity(density,f):
    return (6*math.pi**2*density/f)**(1./3)

#########################################################
#Function that takes in a Fermi momentum and returns the# 
#corresponding density given a degeneracy factor f      #
#########################################################
@jit
def density(kF,f):
    return f*kF**3./(6*math.pi**2)

#######################################################
#Returns te scalar density for a given Fermi momentum,#
#effective mass and degeneracy factor f               #
#######################################################
@jit
def scalarDensity(kF,gSigma,mStar,f):
    x = kF/mStar
    return f/4.*mStar**3*math.pi**(-2)*(np.sqrt(x**2+1.)*x-np.arcsinh(x))


#######################################################
#Returns the effective mass for a given Fermi momentum#
#and degeneracy factor f                              #
#######################################################
def effectiveMass(kF,gSigma,f):
    a = lambda mStar: 1.-gSigma**2*mSigma**(-2)*scalarDensity(kF,gSigma,mStar,f)-mStar
    if(np.sign(a(10**(-16)))==np.sign(a(1.))):
        return 10**(-16)
    return op.brentq(a,10**(-16),1.)

######################################################################
#Takes in a maximum value for the Fermi momenta kFmax, and the length# 
#N1 requested. Returns one vector containing the Fermi momenta evenly#
#spaced between 0 and kFmax and two vectors containing the           #
#corresponding effective masses for degeneracy factor f=2 and f=4    #
######################################################################
@jit
def massVec(kFMax,gSigma,N1):
    kFvec = np.zeros(N1+1)
    kFvec = np.linspace(10**(-16),kFMax,N1)
    mStarVec2 = np.zeros(N1)
    mStarVec4 = np.zeros(N1)
    for i in range(N1):
        mStarVec2[i] = effectiveMass(kFvec[i],gSigma,2.)
        mStarVec4[i] = effectiveMass(kFvec[i],gSigma,4.)
    return kFvec,[mStarVec2,mStarVec4]

################################################################
#Takes in a Fermi momenta, effective mass and degeneracy factor# 
#and returns the pressure                                      #
################################################################ 
@jit
def pressure(kF,gSigma,gOmega,mStar,f):
    x = kF/mStar
    a = 0.5*gSigma**2*mSigma**(-2)*scalarDensity(kF,gSigma,mStar,f)**2
    b = 0.5*gOmega**2*mOmega**(-2)*density(kF,f)**2
    c = f*(48*math.pi**2)**(-1)*mStar**4
    d = np.sqrt(x**2+1.)*(2*x**3-3*x)+3*np.arcsinh(x)
    return -a+b+c*d

#################################################################
#Takes in one vector countaining all the Fermi momenta and      #
#two containing the effective masses for each degenarcy factor, #
#as well as the length of the vectors. Returns the corresponding#
#pressures for each degenracy factor                            #
################################################################# 
@jit
def pressureVec(kFvec,gSigma,gOmega,mStarVec,N1):
    Pvec2 = np.zeros(N1)
    Pvec4 = np.zeros(N1)
    mStarVec2 = mStarVec[0]
    mStarVec4 = mStarVec[1]
    for i in range(N1):
        Pvec2[i] = pressure(kFvec[i],gSigma,gOmega,mStarVec2[i],2.)
        Pvec4[i] = pressure(kFvec[i],gSigma,gOmega,mStarVec4[i],4.)
    return [Pvec2,Pvec4]

#############################################################
#Takes in a Fermi momenta and effective mass and returns the# 
#energy density for a given degeneracy factor               #
#############################################################
@jit
def energyDensity(kF,gSigma,gOmega,mStar,f):
    x = kF/mStar
    a = 0.5*gSigma**2*mSigma**(-2)*scalarDensity(kF,gSigma,mStar,f)**2
    b = 0.5*gOmega**2*mOmega**(-2)*density(kF,f)**2
    c = f*(16*math.pi**2)**(-1)*mStar**4
    d = np.sqrt(x**2+1.)*(2*x**3+x)-np.arcsinh(x)
    return a+b+c*d   

####################################################################
#Same as the function pressureVec, except that the vectors returned#
#are energy densities                                              #
####################################################################
@jit
def energyDensityVec(kFvec,gSigma,gOmega,mStarVec,N1):
    epsilonVec2 = np.zeros(N1)
    epsilonVec4 = np.zeros(N1)
    mStarVec2 = mStarVec[0]
    mStarVec4 = mStarVec[1]
    for i in range(N1):
        epsilonVec2[i] = energyDensity(kFvec[i],gSigma,gOmega,mStarVec2[i],2.)
        epsilonVec4[i] = energyDensity(kFvec[i],gSigma,gOmega,mStarVec4[i],4.)
    return [epsilonVec2,epsilonVec4] 

#######################################################
#Same as the function pressure, except that it returns#
#the pressure for a free Fermi gas                    #
#######################################################
@jit
def fermiGasPressure(kF,mStar,f):
    a = f/(48*math.pi**2)
    b = np.sqrt(kF**2+mStar**2)
    c = 2*kF**3-3*mStar**2*kF
    return a*(b*c+3*mStar**4*np.log((kF+b)/mStar))

##########################################################
#Same as the function pressureVec, except that it returns#
#the pressure vecotrs for a free Fermi gas               #
##########################################################
@jit
def fermiGasPressureVec(kFvec,N1):
    fgPvec2 = fermiGasPressure(kFvec,np.ones(N1),2)
    fgPvec4 = fermiGasPressure(kFvec,np.ones(N1),4)
    return fgPvec2,fgPvec4
    
############################################################
#Same as the function energyDensity, except that it returns#
#the energy density for a free Fermi gas                   #
############################################################
def fermiGasEnergyDensity(kF,mStar,f):
    a = f/(48*math.pi**2)
    b = np.sqrt(kF**2+mStar**2)
    c = 6*kF**3+3*mStar**2*kF
    return a*(b*c-3*mStar**4*np.log((kF+b)/mStar))

###############################################################
#Same as the function energyDensityVec, except that it returns#
#the energy density vectors for a free Fermi gas              #
###############################################################
@jit
def fermiGasEnergyDensityVec(kFvec,N1):
    fgEpsilonVec2 = fermiGasEnergyDensity(kFvec,np.ones(N1),2)
    fgEpsilonVec4 = fermiGasEnergyDensity(kFvec,np.ones(N1),4)
    return fgEpsilonVec2,fgEpsilonVec4

#################################################################
#Function that takes in pressure vectors for the sigma-omega EoS#
#and the free Fermi gas EoS, and combines them at the point     #
#where they conside                                             #
#################################################################
@jit
def combineFermiSigmaOmega(Pvec,fgPvec,N1):
    print(N1)
    index = 0
    for i in range(N1-1):
        if(Pvec[i]>10**(-5)):
            if(np.sign(Pvec[i]-fgPvec[i])!=np.sign(Pvec[i+1]-fgPvec[i+1])):
                index = i+1
                break
    PvecCorr = np.zeros(N1)
    PvecCorr[0:index] = fgPvec[0:index]
    PvecCorr[index:N1] = Pvec[index:N1]
    return PvecCorr


##############################################################################
#Takes in a maximal Fermi momentum and the length of the vector requested    #
#Returns all vectors needed to calculate the mass-radius relations (and more)#
##############################################################################
@jit
def createVectors(kFMax,N1,gSigma,gOmega):
    kFvec,mStarVec = massVec(kFMax,gSigma,N1)
    Pvec = pressureVec(kFvec,gSigma,gOmega,mStarVec,N1)
    epsilonVec = energyDensityVec(kFvec,gSigma,gOmega,mStarVec,N1)
    fgPvec = fermiGasPressureVec(kFvec,N1)
    PvecCorr2 = combineFermiSigmaOmega(Pvec[0],fgPvec[0],N1)
    PvecCorr4 = combineFermiSigmaOmega(Pvec[1],fgPvec[1],N1)
    PvecCorr = [PvecCorr2,PvecCorr4]
    fgEpsilonVec = fermiGasEnergyDensityVec(kFvec,N1)
    
    gSigma = 10.7522
    gOmega = 15.8533
    kFvec,mStarVec = massVec(kFMax,gSigma,N1)
    PvecCoupl = pressureVec(kFvec,gSigma,gOmega,mStarVec,N1)
    epsilonVecCoupl = energyDensityVec(kFvec,gSigma,gOmega,mStarVec,N1)
    
    return kFvec,Pvec,epsilonVec,fgEpsilonVec,fgPvec,PvecCorr,epsilonVecCoupl,PvecCoupl

################################################################
#Takes in two vectors containing the pressure and enery density#
#values and creates plots of the EoS                           #
################################################################
def plotEoS(Pvec,epsilonVec):
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    epsilonVec2 = epsilonVec[0]
    epsilonVec4 = epsilonVec[1]
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$\\bar{\epsilon}$', fontsize = 15)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 15)
    ax.text(-0.00004,-0.4*10**(-7),'E',fontsize = 14)
    ax.text(0.00006,0.7*10**(-7),'D',fontsize = 14)
    ax.text(1.7*10**(-4),-0.4*10**(-7),'C',fontsize = 14)
    ax.text(0.00058,-1.45*10**(-6),'B',fontsize = 14)
    ax.text(0.00082,-0.4*10**(-7),'A',fontsize=14)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim(-0.00005,0.001)
    plt.ylim(-0.0000015,0.0000007)
    plt.plot(epsilonVec2,Pvec2,'r')

    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$\\bar{\epsilon}$', fontsize = 14)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 15)
    ax.text(-6*10**(-5),-5*10**(-7),'E',fontsize = 14)
    ax.text(0.00002,5*10**(-7),'D',fontsize = 14)
    ax.text(1.5*10**(-4),-5*10**(-7),'C',fontsize = 14)
    ax.text(0.001,-1.45*10**(-5),'B',fontsize = 14)
    ax.text(0.00143,-5*10**(-7),'A',fontsize=14)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim(-7*10**(-5),0.0016)
    plt.ylim(-0.000016,0.00004)
    plt.plot(epsilonVec4,Pvec4,'b')

############################################################
#Takes in the "normal" sigma omega EoS and the one where we# 
#have altered the couplings. Then plots them together      #
############################################################
def plotEoSDifferentCouplings(PvecCoupl,epsilonVecCoupl,Pvec,epsilonVec):
    PvecCoupl2 = PvecCoupl[0]
    PvecCoupl4 = PvecCoupl[1]
    epsilonVecCoupl2 = epsilonVecCoupl[0]
    epsilonVecCoupl4 = epsilonVecCoupl[1]
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    epsilonVec2 = epsilonVec[0]
    epsilonVec4 = epsilonVec[1]
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$\\bar{\epsilon}$', fontsize = 15)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 15)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim(0,1.2*10**(-3))
    plt.ylim(-0.25*10**(-5),1.5*10**(-5))
    plt.plot(epsilonVec2,Pvec2,'r',label='original couplings')
    plt.plot(epsilonVecCoupl2,PvecCoupl2,'y',label = 'altered couplings')
    plt.legend()

    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.05, -0.02)
    ax.set_xlabel('$\\bar{\epsilon}$', fontsize = 14)
    ax.yaxis.set_label_coords(-0.05, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 15)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim(0,1.7*10**(-3))
    plt.ylim(-1.3*10**(-5),8*10**(-5))
    plt.plot(epsilonVec4,Pvec4,'b',label='original couplings')
    plt.plot(epsilonVecCoupl4,PvecCoupl4,'g',label = 'altered couplings')
    plt.legend()

########################################################
#Plots pressure as a funciton of inverse energy density#
########################################################
def plotPV(kFvec,Pvec,epsilonVec):
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    epsilonVec2 = epsilonVec[0]
    epsilonVec4 = epsilonVec[1]
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('log($1/\\bar{\epsilon}$)', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$\\bar{P}(1/\\bar{\epsilon})$', rotation='horizontal',fontsize = 14)
    plt.xlim(2,10)
    plt.ylim(-0.000002,0.000002)
    ax.text(2400,-0.4*10**(-7),'C',fontsize = 14)
    ax.text(1300,-3.5*10**(-6),'B',fontsize = 14)
    ax.text(500,-0.75*10**(-6),'A',fontsize=14)
    plt.plot(np.log10(1./epsilonVec2),Pvec2,'r')
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.text(2.900,-1.*10**(-7),'A',fontsize = 14)
    ax.text(3.1700,-15*10**(-7),'B',fontsize = 14)
    ax.text(3.55,-1.*10**(-7),'C',fontsize=14)
    ax.text(4,6*10**(-8),'D',fontsize=14)
    ax.text(9.8,2*10**(-8),'E',fontsize=14)

    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('log($1/\\bar{\epsilon}$)', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{\epsilon})$', rotation='horizontal',fontsize = 14)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ticklabel_format(style='sci' ,axis='y', scilimits=(0,0))
    ax.text(2.600,-4*10**(-7),'A',fontsize = 14)
    ax.text(2.9500,-12.2*10**(-6),'B',fontsize = 14)
    ax.text(3.45,-4*10**(-7),'C',fontsize=14)
    ax.text(4.7,2*10**(-7),'D',fontsize=14)
    ax.text(9.8,2*10**(-7),'E',fontsize=14)
    plt.xlim(2,10)
    plt.ylim(-1.25*10**(-5),0.000005)
    plt.plot(np.log10(1./epsilonVec4),Pvec4,'b')

#######################################################
#Plots the combined free Fermi gas and sigma omega EoS#
#######################################################
def plotCorrectedPressure(kFvec,PvecCorr,Pvec):
    PvecCorr2 = PvecCorr[0]
    PvecCorr4 = PvecCorr[1]
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('$\\bar{k}_F$', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{k}_F)$', rotation='horizontal',fontsize = 14)
    plt.plot(kFvec,Pvec2,'y',label = '$\sigma-\omega$')
    plt.plot(kFvec,PvecCorr2,'r',label = 'corrected')
    plt.legend()
    plt.xlim(0,0.4)
    plt.ylim(-0.1*10**(-4),1.5*10**(-4))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('$\\bar{k}_F$', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$\\bar{P}(\\bar{k}_F)$', rotation='horizontal',fontsize = 14)
    plt.plot(kFvec,Pvec4,'g',label = '$\sigma-\omega$')
    plt.plot(kFvec,PvecCorr4,'b',label = 'corrected')
    plt.legend()
    plt.xlim(0,0.4)
    plt.ylim(-0.2*10**(-4),1.5*10**(-4))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    
##########################################################
#Takes in a pressure P and a vector with pressure values.#
#Returns the index of the vector closest to P            #
##########################################################
@jit
def findIndex(P,Pvec,N1):
    P2 = P[0]
    P4 = P[1]
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    index2 = np.zeros(3,int)
    index4 = np.zeros(3,int)
    a2 = 0
    a4 = 0
    tolerance2 = 100
    tolerance4 = 100
    for i in range(N1-1):

        if(abs(Pvec2[i]-Pvec2[i+1])<tolerance2):
            tolerance2 = abs(Pvec2[i]-Pvec2[i+1])
        
        if(abs(Pvec4[i]-Pvec4[i+1])<tolerance4):
            tolerance4 = abs(Pvec4[i]-Pvec4[i+1])

    for i in range(N1-1):
        if((np.sign(P2-Pvec2[i])!=np.sign(P2-Pvec2[i+1])) or abs(Pvec2[i]-P2)<tolerance2):
            index2[a2] = i+1  
            a2+=1
        if((np.sign(P4-Pvec4[i])!=np.sign(P4-Pvec4[i+1])) or abs(Pvec4[i]-P4)<tolerance4):
            index4[a4] = i+1
            a4+=1
            
    flag2 = True
    flag4 = True
    
    if(a2<3):
        flag2 = False
                   
    if(a4<3):
        flag4 = False
    return [index2,index4],[flag2,flag4]
 
###################################################################
#Returns the area under the curve of the pressure vs invese energy#
#density between index start and stop for the energy density      #
### ###############################################################
def area(epsilonVec,Pvec,start,stop):
    return integrate.simps(1./epsilonVec[start:stop],Pvec[start:stop])


################################
###                          ###
###   Mass-radii relations   ###   
###                          ###
################################
       
###########################################################
#The derivative of the mass M with respect to the radius r#
#for arbitrary relativity                                 #
###########################################################
@jit
def dMdr(r,epsilon):
    return beta*epsilon*r**2

###############################################################
#The derivative of the pressure P with respect to the radius r#
#for arbitrary relativity                                     #
###############################################################
@jit
def dPdrTOV(r,P,M,epsilon):
    return-R0*r**(-2.)*(epsilon+P)*(M+beta*P*r**3)*(1.-2.*R0*M*r**(-1))**(-1)


###################################################################
#Function that creates the coefficients in the Runge Kutta routine# 
###################################################################
@jit
def k(r,P,M,epsilon,h):
    kP1 = dPdrTOV(r,P,M,epsilon)
    kM1 = dMdr(r,epsilon)
    kP2 = dPdrTOV(r+h/2,P+h/2*kP1,M+h/2*kM1,epsilon)
    kM2 = dMdr(r+h/2,epsilon)
    kP3 = dPdrTOV(r+h/2,P+h/2*kP2,M+h/2*kM2,epsilon)
    kM3 = dMdr(r+h/2,epsilon)
    kP4 = dPdrTOV(r+h/2,P+h*kP3,M+h*kM3,epsilon)
    kM4 = dMdr(r+h,epsilon)
    return kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4

    

################################################################
#Function that returns the mass and radius of a star with given# 
#central pressure Pc for both degeneracy factors. It also      #
#returns two flags to indicate if the routine converged        # 
################################################################
@jit
def results(Pc,h,nMax,EoS2,EoS4):
    P2 = Pc
    M2 = 0.
    r2 = 0.
    flag2=True
    P4 = Pc
    M4 = 0.
    r4 = 0.
    flag4=True
    tolerance = 10**(-7)
    for i in range(nMax):
        if(flag2==True):
            Memoryr2 = r2
            MemoryM2 = M2
            r2 = r2+h
            epsilon2 = EoS2(P2)
            kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4 = k(r2,P2,M2,epsilon2,h)
            P2 = P2+h/6*(kP1+2*kP2+2*kP3+kP4)
            M2 = M2+h/6*(kM1+2*kM2+2*kM3+kM4)
            if(np.float(np.real(P2))<=0. or (((M2-MemoryM2)/(M2+MemoryM2))**2<tolerance**2 and M2!=0 and MemoryM2!=0)):
                r2 = Memoryr2
                flag2 = False
                           
        if(flag4==True): 
            Memoryr4 = r4
            MemoryM4 = M4
            r4 = r4+h
            epsilon4 = EoS4(P4)
            kP1,kP2,kP3,kP4,kM1,kM2,kM3,kM4 = k(r4,P4,M4,epsilon4,h)
            P4 = P4+h/6*(kP1+2*kP2+2*kP3+kP4)
            M4 = M4+h/6*(kM1+2*kM2+2*kM3+kM4)
            if(np.float(np.real(P4))<=0. or (((M4-MemoryM4)/(M4+MemoryM4))**2<tolerance**2 and M4!=0 and MemoryM4!=0)):
                r4 = Memoryr4
                flag4 = False
        
        if((flag2 == False) and (flag4 == False)):
            break
        
        
    if(flag2 == True):
            print("Maximum number of iterations reached")
            print("for TOV-equation with Pc = %.8f"%Pc)
            print("with arbitrary relativity and degenerassy 2\n")
    
    if(flag4 == True):
            print("Maximum number of iterations reached")
            print("for TOV-equation with Pc = %.8f"%Pc)
            print("with arbitrary relativity and degenerassy 4\n")
    return r2, np.float(np.real(M2)),r4,np.float(np.real(M4)),flag2,flag4
  
        
#####################################################################
#Parametrises the mass-radus relation from central pressures between# 
#PcMin and PcMax                                                    #
#####################################################################
@jit    
def paramterising(parameters,epsilonVec,Pvec):
    PcMin = parameters[0]
    PcMax = parameters[1]
    N2 = parameters[2]
    h = parameters[3]
    nMax = parameters[4]
    
    epsilonVec2 = epsilonVec[0]
    epsilonVec4 = epsilonVec[1]
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    
    Pc = PcMin
    N1 = len(Pvec2)
    if(PcMax>Pvec2[N1-1]):
        print("PcMax changed from ",PcMax," to ", Pvec2[N1-1])
        PcMax = Pvec2[N1-1]
    if(PcMax>Pvec4[N1-1]):
        print("PcMax changed from ",PcMax," to ", Pvec4[N1-1])
        PcMax = Pvec4[N1-1]
    const = (np.float(PcMax)/PcMin)**(1./N2)
    R2 = np.zeros(N2,np.double)
    M2 = np.zeros(N2,np.double)
    R4 = np.zeros(N2,np.double)
    M4 = np.zeros(N2,np.double)

    EoS2 = interpolate.interp1d(Pvec2,epsilonVec2)
    EoS4 = interpolate.interp1d(Pvec4,epsilonVec4)
    
    for i in range(N2):
        print(i)
        R2[i],M2[i],R4[i],M4[i],flag2,flag4 = results(Pc,h,nMax,EoS2,EoS4)
        if((flag2==True) or (flag4 == True)):
            print(R2[i],R4[i])
            R2 = R2[0:i]
            M2 = M2[0:i]
            R4 = R4[0:i]
            M4 = M4[0:i]
            N = i
            break
        Pc = Pc*const
    return R2,M2,R4,M4,N2
       
######################################
#Writes mass-radius relations to file#
######################################
def writeResultsToFile(parameters,epsilonVec,Pvec,filename):

    R2,M2,R4,M4,N2 = paramterising(parameters,epsilonVec,Pvec)
    
    f2 = open(filename+'2','w')
    f2.write(str(N2)+'\n')
    for i in range(N2):
        a = str(R2[i])
        b = str(M2[i])
        f2.write(a+" "+b+"\n")
    f2.close()
    
    f4 = open(filename+'4','w')
    f4.write(str(N2)+'\n')
    for i in range(N2):
        a = str(R4[i])
        b = str(M4[i])
        f4.write(a+" "+b+"\n")
    f4.close()
    
    return   

##########################################
#Writes all mass-radius relations to file#
##########################################
def writeAllResultsToFile(parameters,epsilonVec,Pvec,PvecCorr,PvecCoupl,filenameVec):
    writeResultsToFile(parameters,epsilonVec,Pvec,filenameVec[0])
    writeResultsToFile(parameters,epsilonVec,PvecCorr,filenameVec[1])
    writeResultsToFile(parameters,epsilonVec,PvecCoupl,filenameVec[2])

    
    

######################################
#Reads mass-radius relation from file#
######################################
def readResultsFromFile(filename):
    f2 = open(filename+'2','r')
    N2 = int(f2.readline())
    R2 = np.zeros(N,np.double)
    M2 = np.zeros(N,np.double)
    i = 0
    data = f2.readlines()
    for line in data:
        numbers = line.split()
        R2[i] = numbers[0]
        M2[i] = numbers[1]
        i = i+1 
    f2.close()
    
    f4 = open(filename+'4','r')
    N2 = int(f4.readline())
    R4 = np.zeros(N,np.double)
    M4 = np.zeros(N,np.double)
    i = 0
    data = f4.readlines()
    for line in data:
        numbers = line.split()
        R4[i] = numbers[0]
        M4[i] = numbers[1]
        i = i+1 
    f4.close()
    
    return [R2,R4],[M2,M4],N2

####################################################################
#Plots the mass-radius relation for a given mass and radius vector.#
#Colour is a vector with length 2 that decides the colour of the   #
#curves for f=2 and f=4.                                           #
####################################################################
def plotMassRadiusRelations(R,M,N,colour,Label):
    R2 = R[0]
    R4 = R[1]
    M2 = M[0]
    M4 = M[1]
    R2stable = R2[0:list(M2).index(max(M2))+1]
    R2unstable = R2[list(M2).index(max(M2)):N]
    M2stable = M2[0:list(M2).index(max(M2))+1]
    M2unstable = M2[list(M2).index(max(M2)):N]
    R4stable = R4[0:list(M4).index(max(M4))+1]
    R4unstable = R4[list(M4).index(max(M4)):N]
    M4stable = M4[0:list(M4).index(max(M4))+1]
    M4unstable = M4[list(M4).index(max(M4)):N]
    
    print('Maximum mass for '+colour[0]+' is',max(M2), 'solar masses with radius', R2[list(M2).index(max(M2))],'km')
    print('Maximum mass for '+colour[1]+' is',max(M4), 'solar masses with radius', R2[list(M4).index(max(M4))],'km')
    
    plt.figure()
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('$R$[km]', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$M/M_\odot$', rotation='horizontal',fontsize = 14)
    plt.plot(R2stable,M2stable,colour[0],label=Label[0])
    plt.plot(R2unstable,M2unstable,colour[0]+'--')
    plt.plot(R4stable,M4stable,colour[1],label=Label[1])
    plt.plot(R4unstable,M4unstable,colour[1]+'--')
    plt.xlim(3,29)
    plt.legend()
    
########################################################
#Function that plots all relevant mass-radius relations#
########################################################
def plotAllMassRadiusRelationsFromFile(filenameVec,colourVec):
    R,M,N = readResultsFromFile(filenameVec[0])
    Rcorr,Mcorr,N = readResultsFromFile(filenameVec[1])
    Rcoupl,Mcoupl,N = readResultsFromFile(filenameVec[2])
    
    Label = ['f=2','f=4']
    plotMassRadiusRelations(R,M,N,colourVec[0],Label)
    plotMassRadiusRelations(Rcorr,Mcorr,N,colourVec[1],Label)
    plotMassRadiusRelations(Rcoupl,Mcoupl,N,colourVec[2],Label)
    
    R2 = R[0]
    R4 = R[1]
    M2 = M[0]
    M4 = M[1]
    
    Rcorr2 = Rcorr[0]
    Rcorr4 = Rcorr[1]
    Mcorr2 = Mcorr[0]
    Mcorr4 = Mcorr[1]
    
    Rcoupl2 = Rcoupl[0]
    Rcoupl4 = Rcoupl[1]
    Mcoupl2 = Mcoupl[0]
    Mcoupl4 = Mcoupl[1]
    
    R2stable = R2[0:list(M2).index(max(M2))+1]
    R2unstable = R2[list(M2).index(max(M2)):N]
    M2stable = M2[0:list(M2).index(max(M2))+1]
    M2unstable = M2[list(M2).index(max(M2)):N]
    
    Rcorr2stable = Rcorr2[0:list(Mcorr2).index(max(Mcorr2))+1]
    Rcorr2unstable = Rcorr2[list(Mcorr2).index(max(Mcorr2)):N]
    Mcorr2stable = Mcorr2[0:list(Mcorr2).index(max(Mcorr2))+1]
    Mcorr2unstable = Mcorr2[list(Mcorr2).index(max(Mcorr2)):N]
    
    Rcoupl2stable = Rcoupl2[0:list(Mcoupl2).index(max(Mcoupl2))+1]
    Rcoupl2unstable = Rcoupl2[list(Mcoupl2).index(max(Mcoupl2)):N]
    Mcoupl2stable = Mcoupl2[0:list(Mcoupl2).index(max(Mcoupl2))+1]
    Mcoupl2unstable = Mcoupl2[list(Mcoupl2).index(max(Mcoupl2)):N]
    
    R4stable = R4[0:list(M4).index(max(M4))+1]
    R4unstable = R4[list(M4).index(max(M4)):N]
    M4stable = M4[0:list(M4).index(max(M4))+1]
    M4unstable = M4[list(M4).index(max(M4)):N]
    
    Rcorr4stable = Rcorr4[0:list(Mcorr4).index(max(Mcorr4))+1]
    Rcorr4unstable = Rcorr4[list(Mcorr4).index(max(Mcorr4)):N]
    Mcorr4stable = Mcorr4[0:list(Mcorr4).index(max(Mcorr4))+1]
    Mcorr4unstable = Mcorr4[list(Mcorr4).index(max(Mcorr4)):N]
    
    Rcoupl4stable = Rcoupl4[0:list(Mcoupl4).index(max(Mcoupl4))+1]
    Rcoupl4unstable = Rcoupl4[list(Mcoupl4).index(max(Mcoupl4)):N]
    Mcoupl4stable = Mcoupl4[0:list(Mcoupl4).index(max(Mcoupl4))+1]
    Mcoupl4unstable = Mcoupl4[list(Mcoupl4).index(max(Mcoupl4)):N]
    
    colour = colourVec[0]
    colourCorr = colourVec[1]
    colourCoupl = colourVec[2]
    
    plt.figure()
    plt.plot(R2stable,M2stable,colour[0],label = '$\sigma-\omega$, bounded, f=2')
    plt.plot(R2unstable,M2unstable,colour[0]+'--')
    plt.plot(R4stable,M4stable,colour[1],label = '$\sigma-\omega$, bounded, f=4')
    plt.plot(R4unstable,M4unstable,colour[1]+'--')
    plt.plot(Rcoupl2stable,Mcoupl2stable,colourCoupl[0],label = 'New Couplings $\sigma-\omega$, f=2')
    plt.plot(Rcoupl2unstable,Mcoupl2unstable,colourCoupl[0]+'--')
    plt.plot(Rcoupl4stable,Mcoupl4stable,colourCoupl[1],label = 'New Couplings $\sigma-\omega$, f=4')
    plt.plot(Rcoupl4unstable,Mcoupl4unstable,colourCoupl[1]+'--')
    plt.plot(Rcorr2stable,Mcorr2stable,colourCorr[0],label = 'Fermi-$\sigma-\omega$, f=2')
    plt.plot(Rcorr2unstable,Mcorr2unstable,colourCorr[0]+'--')
    plt.plot(Rcorr4stable,Mcorr4stable,colourCorr[1],label = 'Fermi-$\sigma-\omega$, f=4')
    plt.plot(Rcorr4unstable,Mcorr4unstable,colourCorr[1]+'--')
    plt.xlim(3,29)


    plt.legend()
    
    ax = plt.gca()
    ax.xaxis.set_label_coords(1.07, -0.02)
    ax.set_xlabel('$R$[km]', fontsize = 14)
    ax.yaxis.set_label_coords(-0.06, 1.03)
    ax.set_ylabel('$M/M_\odot$', rotation='horizontal',fontsize = 14)

    
##########################################################################
#Calculates the coupling constants within an interval divided in N pieces#
##########################################################################
@jit
def couplingConstants(gSigmaRange,gOmegaRange,N):
    satDens = fmToMeV(0.153,-3)/939**3
    bindingEnergy = -16.3/939
    gSigma = gSigmaRange[0]
    gOmega = gOmegaRange[0]
    f=4
    hSigma = (gSigmaRange[1]-gSigmaRange[0])/(N-1)
    hOmega = (gOmegaRange[1]-gSigmaRange[0])/(N-1)
    kFsat = kFfromDensity(satDens,f)
    bestFit = 100.
    bestGsigma = 0.
    bestGomega = 0.
    h = 10**(-10)
    rhoPluss = density(kFsat+h,f)
    rhoMinus = density(kFsat-h,f)
    drho = rhoPluss-rhoMinus
    
    for i in range(N):
        for j in range(N):
            mStar = effectiveMass(kFsat,gSigma,f)
            epsilon = energyDensity(kFsat,gSigma,gOmega,mStar,f)
            mStarPluss = effectiveMass(kFsat+h,gSigma,f)
            epsilonPluss = energyDensity(kFsat+h,gSigma,gOmega,mStarPluss,f)
            mStarMinus = effectiveMass(kFsat-h,gSigma,f)
            epsilonMinus = energyDensity(kFsat-h,gSigma,gOmega,mStarMinus,f)
            
            testBindingEnergy = epsilon/satDens-1.
            
            derivativeEpsilon = (epsilonPluss-epsilonMinus)/(drho)
            
            k = 1.-testBindingEnergy/bindingEnergy
            l = 1.-derivativeEpsilon/epsilon*satDens
            test = np.sqrt(k**2+l**2)
            
            if(test<bestFit):
                bestGsigma = gSigma
                bestGomega = gOmega 
                bestFit = test
                
            gOmega += hOmega 
             
        gOmega = gOmegaRange[0] 
        gSigma += hSigma 
        
    return bestGsigma,bestGomega
    
##############################################################
#N decides how many pieces we divide the interval [0,200] in.#
#Returns the obtained coupling constants                     #
##############################################################
@jit
def couplings(N):
    gSigmaRange = [0,200]
    gOmegaRange = [0,200]
    gSigma,gOmega = couplingConstants(gSigmaRange,gOmegaRange,N)
    print(gSigma,gOmega)
    gSigmaRange[0] = gSigma-10.
    gSigmaRange[1] = gSigma+10.
    gOmegaRange[0] = gOmega-10.
    gOmegaRange[1] = gOmega+10.
    gSigma,gOmega = couplingConstants(gSigmaRange,gOmegaRange,N)
    print(gSigma,gOmega)
    gSigmaRange[0] = gSigma-1.
    gSigmaRange[1] = gSigma+1.
    gOmegaRange[0] = gOmega-1.
    gOmegaRange[1] = gOmega+1.
    gSigma,gOmega = couplingConstants(gSigmaRange,gOmegaRange,N)
    print(gSigma,gOmega)
    gSigmaRange[0] = gSigma-0.1
    gSigmaRange[1] = gSigma+0.1
    gOmegaRange[0] = gOmega-0.1
    gOmegaRange[1] = gOmega+0.1
    gSigma,gOmega = couplingConstants(gSigmaRange,gOmegaRange,N)
    return gSigma,gOmega





##################################################################################
#Plots the binding energy. choose nc = 'n' for normal view, and nc = 'z' for zoom#
##################################################################################
def plotBindingEnergy(kFvec,epsilonVec,nc):
    plt.figure()
    kFvecMeV = 939*kFvec
    epsilonVecMeV2 = epsilonVec[0]*939**4
    epsilonVecMeV4 = epsilonVec[1]*939**4
    rho2 = density(kFvecMeV,2)
    rho4 = density(kFvecMeV,4)
    m = 939*np.ones(len(kFvec))
    B2 = epsilonVecMeV2/rho2-m
    B4 = epsilonVecMeV4/rho4-m
    B2[0] = 0
    B4[0] = 0
    zero = np.zeros(len(kFvec))
    rho2 = MeVtoFm(density(kFvecMeV,2),3)
    rho4 = MeVtoFm(density(kFvecMeV,4),3)
    plt.plot(rho4,zero,'g--')
    plt.plot(rho2,B2,'r',label = 'f=2')
    plt.plot(rho4,B4,'b',label = 'f=4')
    
    if(nc=='n'):
        plt.xlim(0,0.5)
        plt.ylim(-20,100)
    elif(nc=='z'):
        plt.xlim(0,0.3)
        plt.ylim(-1,4)
    
    ax = plt.gca()
    ax.xaxis.set_label_coords(0.9, -0.07)
    ax.set_xlabel('$\\rho$ [fm$^{-3}$]', fontsize = 14)
    ax.yaxis.set_label_coords(-0, 1.04)
    ax.set_ylabel('$B(\\rho)$ [MeV]', rotation='horizontal',fontsize = 14)
    plt.legend()

    
    
#####################################################
#Function that plots all the figures given in thesis#
#####################################################
def plotEverything(kFvec,Pvec,epsilonVec,fgEpsilonVec,fgPvec,PvecCorr,epsilonVecCoupl,PvecCoupl,filenameVec):
    colour = ['r','b']
    colorCorr = ['y','g']
    colourCoupl = ['m','k']
    colourVec = [colour,colorCorr,colourCoupl]
    plotEoS(Pvec,epsilonVec)
    plotEoSDifferentCouplings(PvecCoupl,epsilonVecCoupl,Pvec,epsilonVec)
    plotPV(kFvec,Pvec,epsilonVec)
    plotCorrectedPressure(kFvec,PvecCorr,Pvec)
    plotAllMassRadiusRelationsFromFile(filenameVec,colourVec)
    plotBindingEnergy(kFvec,epsilonVec,'n')
    plotBindingEnergy(kFvec,epsilonVec,'z')



######################################################################
#Function that checks if the Maxwell construction is possible for    #
#a given EoS. Returns a string containg a message that can be printed#
######################################################################
def isMaxwellPossible(epsilonVec,Pvec,N1):
    index,flags = findIndex([0,0],Pvec,N1)
    text = ''
    if(flags[0] == False): #Checks if the function is strictly increasing#
        text+='Maxwell construction not nercecary for f=2. Function is strictly increasing!\n'
    if(flags[1] == 0):
        text+='Maxwell construction not nercecary for f=4. Function is strictly increasing!\n'
    index2 = index[0]
    index4 = index[1]
    epsilonVec2 = epsilonVec[0]
    epsilonVec4 = epsilonVec[1]
    Pvec2 = Pvec[0]
    Pvec4 = Pvec[1]
    
    if(flags[0] == True):
        firstArea2 = area(epsilonVec2,Pvec2,index2[0],index2[1])
        secondArea2 = area(epsilonVec2,Pvec2,index2[1],index2[2])
        if(abs(firstArea2)<abs(secondArea2)):
            text+='Maxwell construction not possible for f=2\n'
        else:
            text+='Maxwell construction possible for f=2\n'
    if(flags[1] == True):
        firstArea4 = area(epsilonVec4,Pvec4,index4[0],index4[1])
        secondArea4 = area(epsilonVec4,Pvec4,index4[1],index4[2])
        if(abs(firstArea4)<abs(secondArea4)):
            text+='Maxwell construction not possible for f=4\n'
        else:
            text+='Maxwell construction possible for f=4\n'
    return text



########################                
###                  ###
###   Progran start  ###
###                  ###
########################


#Defining constants and parameters

mSigma = 0.5857294994675186368     #Mass of sigma divided by nucleon mass
mOmega = 0.8338658146964856230     #Mass of omega divided by nucleon mass

R0 = 1.47              #Half the suns Swarzchild radius
beta = 1.1426          #Parameter found in thesis
kFMax = 2              #Maximum nucleon Fermi momentum used to calculate EoS
N = 1000               #Number of divisions of the interall we search for the coupling constants
N1 = 50000             #Number of points used to calculate EoS
N2 = 300               #Number of points used to calculate mass-radius relation
PcMin = 10**(-10)      #Minimum central pressure used in mass-radius relation
PcMax = 10**(4)        #Maximum central pressure used in mass-radius relation
h = 0.0005             #Step length in km used to calculate the mass-radius relation
nMax = int(100./h)     #Maximum number of iterations before we stop the mass calculations           

filename = 'SigmaOmegaMassRadiRelation1'
filenameCorr = 'SigmaOmegaMassRadiRelationCorr1'
filenameCoupl = 'SigmaOmegaMassRadiRelationCoupl1'

#Create a vector containing all filenames
filenameVec = [filename,filenameCorr,filenameCoupl]

#Create vector containing important parameters
parameters = [PcMin,PcMax,N2,h,nMax,kFMax]

#Finds the coupling constants
gSigma,gOmega = couplings(N)

#The progrm that finds the couplings uses a bit of time. If desired, one
#can just remove the comment below and use the obtained values
#gSigma = 10.94
#gOmega = 13.59

#Creates all vectors
kFvec,Pvec,epsilonVec,fgEpsilonVec,fgPvec,PvecCorr,epsilonVecCoupl,PvecCoupl= createVectors(kFMax,N1,gSigma,gOmega)

#Writes all mass-radius relations to file
writeAllResultsToFile(parameters,epsilonVec,Pvec,PvecCorr,PvecCoupl,filenameVec)

#Print if the maxwell construction is possible
print(isMaxwellPossible(epsilonVec,Pvec,N1))

#Produce all plots given in thesis
plotEverything(kFvec,Pvec,epsilonVec,fgEpsilonVec,fgPvec,PvecCorr,epsilonVecCoupl,PvecCoupl,filenameVec)

#Show plots
plt.show()

#Print the time spent
print("\nTime spent:")
print(time.clock()-t0)